! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: espfilt_mod
!
! Authors:  Brendan Perry - NASA/GSFC (complete rewrite in f95) brendan.h.perry@nasa.gov
!           K.D.Kuntz (NASA/GSFC) and The Johns Hopkins University, kuntz@pha.jhu.edu
!           Steve Snowden - NASA/GSFC (original algorithm in perl)
!
! Description: Generates soft proton contamination-filtered products from
!              event lists.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!-
!
module espfilt_mod

  use dal
  use param 
  use errorHandling
  use gofutils_mod  ! from goflib, formerly in marq_mod
  use esas_mod      ! from goflib, KDK code for SAS-21
  use gofio_mod     ! from goflib, formerly in subsn (SS)
  use, intrinsic :: ieee_arithmetic ! use to weed out nulls/divide by zeroes

  implicit none 

  private

  character(len=*), parameter :: &
       evlisetpar     = "eventfile", &     ! was 'prefix'
       withootpar     = "withoot", &       ! process oot file too?
       ootsetpar      = "ootfile", &       ! PN OOT event list
       methodpar      = "method", &        ! histogram, histo, or ratio
       withsmoothpar  = "withsmoothing", & ! smooth data? 
       smoothfactpar  = "smooth", &        ! smoothing factor
       ratiopar       = "ratio", &         ! flare factor ratio 
       deltpar        = "rangescale", &    ! histogram fit range scale factor
       rangescalepar  = "allowsigma", &    ! Limit in sigma for unflared rates.
       elowpar        = "elow",&           ! low energy channel bound
       ehighpar       = "ehigh",&          ! high energy channel bound
       withbinpar     = "withbinning",&    ! bin the data? 
       binsizepar     = "binsize", &       ! bin size?
       plimpar        = "limits",&         ! clean lc plim parameters 
       keeppar        = "keepinterfiles"   ! keep intermediary files?
  public espfilt

  real(kind=double),allocatable,dimension(:)::kip_x,kip_y
  integer::kip_n

contains 

  !---------------------------------------------------------------------
  subroutine espfilt
    !---------------------------------------------------------------------

    type(DataSetT) :: infovEVLI_S
    type(TableT)   :: infovevents_T
    type(ArrayT)   :: infovEVLI_A
    character(len=2)   :: patt, inst
    character(len=5)   :: instrume, cminchan, cmaxchan
    character(len=6)   :: filter
    character(len=10)  :: method, obsid, expid
    character(len=25)  :: smode
    character(len=256) :: inEVLI_F, inOOT_F, fovimg_F, fovlc_F, corimg_F, &
         corlc_F, corevli_F, annEVLI_F, annimg_F, annlc_F, gtitxt_F
    character(len=1024) :: cornerdef, anndef
    integer :: smoothfactor, binsize, elow, ehigh, elowO, ehighO, &
         i, revolut, nRows, typeF, frmtime
    integer :: status = 0
    logical :: withsmoothing, withbinning, withlong, withoot, meth, &
         specialSmode, keep, warndie, k
    real(kind=single) :: ratio, back, area, expo
    real(kind=double) :: rangescale, delt, plim(2)
    elow  = 0
    ehigh = 0
    !
    ! get the parameters from the par file
    !
    write(errstr,*) "ESPFILT: Retrieving parameters from par file."
    call message(AppMsg,Noisy)

    inEVLI_F = stringParameter(evlisetpar) ! input eventlist from e*chain
    withoot  = booleanParameter(withootpar)
    method   = stringParameter(methodpar)    ! histo, histogram, or ratio
    write(errstr,*) "ESPFILT: filtering method: ",method
    call message(AppMsg,Noisy)
    withsmoothing = booleanParameter(withsmoothpar)
    smoothfactor  = intParameter(smoothfactpar)
    ratio = realParameter(ratiopar)            ! flare factor ratio
    delt  = realParameter(deltpar)             ! former rangescale
    rangescale = realParameter(rangescalepar)  ! new rangescale
    !    withlong   = booleanParameter(withlongpar)
    !    if (withlong) then
    !       if (index(method,'ratio') .gt. 0) then
    !          call warning("noLongnamesWratio",&
    !               "Long (energy based) names inappropriate for ratio method, setting to False")
    !          withlong = .FALSE.
    !       endif
    !    endif
    withlong = .FALSE. ! descoped for SAS-20.1, may bring back later???!!!
    elow  = intParameter(elowpar)
    ehigh = intParameter(ehighpar)
    do i=1,2
       plim(i) = realParameter(plimpar,i-1) ! 0.1, 6.5
    enddo
    if (elow .ge. ehigh) then
       call error("badEnergyChans","Parameter elow must be < ehigh")
    end if
    withbinning = booleanParameter(withbinpar)
    binsize = intParameter(binsizepar)
    keep = booleanParameter(keeppar)
    !
    write(errstr,*) "ESPFILT: Processing eventlist: ",trim(adjustl(inEVLI_F))
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(adjustl(inEVLI_F)),infovEVLI_S,1,status,warndie)
    !
    ! Get critical header keywords from input event list
    !
    status = 0
    warndie = .TRUE.
    typeF = 1 ! 1 = EVLI, 2 = SPEC, 3 = IMG
    call read_header_info(typeF,infovEVLI_S,infovevents_T,infovEVLI_A,&
         inst,obsid,expid,smode,frmtime,filter,revolut,back,area,expo,&
         elowO,ehighO,nRows,warndie)
    write(errstr,*) "OBSID = ",trim(adjustl(obsid)),"EXPID = ",trim(adjustl(expid))
    call message(AppMsg,Always)
    if (withoot) then
       inOOT_F = stringParameter(ootsetpar) ! input OOT evli from epchain
       if (trim(adjustl(inOOT_F)) .eq. 'default') then ! if not explicitly given
          inOOT_F = "pn" // trim(adjustl(expid)) // "-allevoot.fits"
       endif
    end if
    if ( trim(adjustl(filter)) .EQ. 'Closed')  then 
       call error ("noClosed",&
            "Sorry, espfilt does not work with Closed data")
    end if
    ! add pattern as a parameter???!!!
    if (inst .eq. 'PN') then
       patt = '4'
    else
       patt = '12'
       if (withoot) then ! nip it in the bud
          call error ("noOOTwMOS",&
               "OOT processing is for EPN only")
       end if
    endif
    !
    ! For PN observations check the SUBMODE keyword in the event file primary 
    ! header, if not 'PrimeFullWindow' or 'PrimeFullWindowExtended' then die 
    ! gracefully and with an explanation
    ! PrimeFullWindow OR PrimeFullWindowExtended
    ! bhp fixed, 03/2021 to address all submodes
    specialSmode = .TRUE.
    if (inst .eq. 'PN') then
       if (index(smode,'Fast') .gt. 0) then
          call error ("noPNFast",&
               "Sorry, espfilt does not work with EPN Fast* SUBMODEs")
       elseif (index(smode,'mall') .gt. 0) then
          call error ("noPNSmallWindow",&
               "Sorry, espfilt does not work with EPN *Small* window SUBMODEs")
       elseif (index(smode,'arge') .gt. 0) then
          call warning ("PNLargeWindow",&
               "EPN PrimeLargeWindow SUBMODE may produce suspect results, but will attempt")
       end if
    end if
    !    
    call release(infovEVLI_S) ! close the ALL EventList
    !
    ! define the corners for evselect calls later
    !
    call defineexpr (method,patt,inst,cornerdef,anndef)
    !
    ! create ALL and corner image and FOV lightcurves
    !
    call makeevimlc (inEVLI_F,method,cornerdef,anndef,inst,expid,&
         specialSmode,revolut,elow,ehigh,patt,fovimg_F,fovlc_F,&
         corevli_F,corimg_F,corlc_F,annEVLI_F,annimg_F,annlc_F,withlong)
    !
    if ( (method .eq. "histo") .or. (method .eq. "histogram") ) then 
       write(errstr,*) "ESPFILT: Processing using the HISTOGRAM method."
       call message(AppMsg,Always)
       call histogram_method (inEVLI_F,cornerdef,elow,ehigh,obsid,expid,&
            revolut,inst,withbinning,withsmoothing,binsize,smoothfactor,patt,&
            specialSmode,rangescale,delt,fovimg_F,fovlc_F,corlc_F,gtitxt_F,&
            withlong,withoot,inOOT_F,plim)
    elseif (method .eq. "ratio") then 
       write(errstr,*) "ESPFILT: Processing using the RATIO method."
       call message(AppMsg,Always)
       call ratio_method (inEVLI_F,cornerdef,anndef,elow,ehigh,obsid,expid,&
            revolut,inst,withbinning,withsmoothing,binsize,smoothfactor,&
            ratio,patt,specialSmode,rangescale,delt,fovimg_F,fovlc_F,corlc_F,&
            annEVLI_F,annlc_F,gtitxt_F,withlong,withoot,inOOT_F,plim)
    else
       ! other methods will eventually go here...
       ! should never happen cuz of parameter switch
       call error("UnsupportedMethod",&
            "Method must be either histogram or ratio.")
    end if
    !
    if (.not.keep) then
       write(errstr,*) "ESPFILT: **** keepinterfiles NOT set"
       call message(AppMsg,Always)
       write(errstr,*) "ESPFILT: **** Deleting these intermediary files: "
       call message(AppMsg,Always)
       if (method .eq. "ratio") then
          write(errstr,*) "Annulus evlist: ",trim(adjustl(annevli_F))
          call message(AppMsg,Always)
          call unlink(trim(adjustl(annimg_F)))
          write(errstr,*) "Annulus image:  ",trim(adjustl(annimg_F))
          call message(AppMsg,Always)
          call unlink(trim(adjustl(annevli_F)))
       end if
       write(errstr,*) "Unfiltered ALL image: ",trim(adjustl(fovimg_F))
       call message(AppMsg,Always)
       call unlink(trim(adjustl(fovimg_F)))
       write(errstr,*) "Corner-only image: ",trim(adjustl(corimg_F))
       call message(AppMsg,Always)
       call unlink(trim(adjustl(corimg_F)))
       write(errstr,*) "GTI text file for gtibuild: ",trim(adjustl(gtitxt_F))
       call message(AppMsg,Always)
       call unlink(trim(adjustl(gtitxt_F)))
    end if

    call message(AppMsg,Verbose)

  end subroutine espfilt
  !
  !--------------------------------------------------
  ! subroutines
  !--------------------------------------------------
  !
  subroutine histogram_method (inEVLI_F,cornerdef,elow,ehigh,obsid,expid,&
       revolut,inst,binning,smoothing,binsize,smoothfactor,patt,specialSmode,&
       rangescale,delt,fovimg_F,fovlc_F,corlc_F,gtitxt_F,withlong,withoot,&
       inOOT_F,plim)
    type(DataSetT) :: set, fovEVLI_S
    type(TableT)   :: tab
    character(len=2) :: inst, patt
    character(len=5) :: suffix
    character(len=6) :: expid
    character(len=10) :: obsid, type
    character(len=18) :: prefix
    character(len=32) :: flagg
    character(len=*), intent(in) :: inEVLI_F
    character(len=256) :: fovimg_F, fovlc_F, corlc_F, gtitxt_F, &
         gtifits_F, fovevlifilt_F, histqdp_F, &
         fovimgfilt_F, corimgfilt_F, corevlifilt_F, inOOT_F, outOOT_F
    character(len=1024) :: cornerdef, command
    integer, intent(in) :: smoothfactor, binsize, revolut
    integer :: method, elow, ehigh, status, typeF

    logical :: specialSmode, warndie = .true. 
    logical, intent(in) :: binning, smoothing, withlong, withoot
    logical :: f_ok = .false.

    real(kind=double) :: rangescale, delt
    real(kind=double) :: ratio = 1.0, norm, width, center, apar(3), &
         h_imme,h_totl,l_imme,l_totl, plim(2)
    apar = (/0.d0,0.d0,0.d0/)
    !
    ! Eliminate QDP file construct, use python-based plotters in SAS-20!!!
    !
    type = "hist"  ! mos1S001-hist.qdp
    suffix = "qdp" ! QDP output file (change to python!!!)
    call make_filename(inst,expid,type,elow,ehigh,suffix,histqdp_F,.FALSE.)
    write(errstr,*) "HISTOGRAM_METHOD: *HISTOGRAM QDP FILE* = "//&
         trim(adjustl(histqdp_F))
    call message(AppMsg,Noisy)
    type = "gti"   ! mos1S001-gti.txt
    suffix = "txt" ! GTI temp text file
    call make_filename(inst,expid,type,elow,ehigh,suffix,gtitxt_F,.FALSE.)
    write(errstr,*) "HISTOGRAM_METHOD: *GTI TXT FILE* = "//&
         trim(adjustl(gtitxt_F))
    call message(AppMsg,Noisy)
    type = "gti"    ! mos1S001-gti.fits
    suffix = "fits" ! GTIBUILD output FITS file
    call make_filename(inst,expid,type,elow,ehigh,suffix,gtifits_F,.FALSE.)
    write(errstr,*) "HISTOGRAM_METHOD: *GTI FITS FILE* = "//&
         trim(adjustl(gtifits_F))
    call message(AppMsg,Noisy)
    !
    ! clean_lc is the engine that creates the gti txt file for use by
    ! gtibuild and the smoothed, binned lc's
    !
    write(errstr,*) "HISTOGRAM_METHOD: Running Clean_LC..."
    call message(AppMsg,Verbose)
    method = 1 ! histogram = 1, ratio = 2
    call clean_lc(obsid,expid,gtitxt_F,fovlc_F,corlc_F,fovlc_F,histqdp_F,&
         binning,smoothing,binsize,smoothfactor,method,ratio,specialSmode,&
         rangescale,delt,norm,width,center,apar,plim,h_imme,h_totl,l_imme,&
         l_totl)

    command = "gtibuild file="// trim(adjustl(gtitxt_F)) //" table="// &
         trim(adjustl(gtifits_F))
    write(errstr,*) "*gtibuild command = "
    call message(AppMsg,Noisy)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Noisy)
    call system(trim(adjustl(command)))
    !
    ! if GTI file is ok, create final, filtered, flare-free eventlist (clean)
    ! Account for MOS1 CCD#4 data after CCD#3 loss???!!!
    !
    if (inst .eq. 'PN') then ! new for SAS-19
       flagg =  '(#XMMEA_EP)'
    else
       !orig:       flagg = "((FLAG & 0x766aa000)==0)"
       flagg = "((FLAG & 0x766aa000)==0)"
    endif
    !
    inquire(file=trim(adjustl(gtifits_F)),exist=f_ok)
    if (f_ok) then
       type = "allevc"
       suffix = "fits" ! OLDNEW: mos1S001-fovevc.fits/mos1S001-allevc.fits
       call make_filename(inst,expid,type,elow,ehigh,suffix,fovevlifilt_F,withlong)
       write(errstr,*) &
            "HISTOGRAM_METHOD: *CLEANED (flare-free) ALL EVENTLIST FILE* = "//&
            trim(adjustl(fovevlifilt_F))
       call message(AppMsg,Always)
       ! removed revolut test below for SAS-19
       command = "evselect table="// trim(adjustl(inEVLI_F)) // &
            " withfilteredset=yes "// &
            "expression='(PATTERN<=" // trim(adjustl(patt)) // ")&&GTI(" // &
            trim(adjustl(gtifits_F)) // ",TIME)&&" // trim(adjustl(flagg)) // &
            "' filteredset="// trim(adjustl(fovevlifilt_F)) // &
            " filtertype=expression keepfilteroutput=yes "// &
            "updateexposure=yes filterexposure=yes"
       write(errstr,*) "*Flare free eventlist evselect command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(fovevlifilt_F,elow,ehigh)
       call addfitpars(fovevlifilt_F,norm,width,center,h_imme,h_totl,l_imme,l_totl)
       !
       type = "allimc"
       suffix = "fits" ! OLDNEW: mos1S001-fovimc.fits/mos1S001-allimc.fits
       call make_filename(inst,expid,type,elow,ehigh,suffix,fovimgfilt_F,withlong)
       write(errstr,*) &
            "HISTOGRAM_METHOD: *CLEANED (flare-free) ALL IMAGE FILE* = "//&
            trim(adjustl(fovimgfilt_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(fovevlifilt_F)) //&
            " withimageset=yes imageset="// trim(adjustl(fovimgfilt_F)) //&
            " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
            "ximagesize=600 yimagesize=600"
       write(errstr,*) "*Flare free ALL image evselect command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(fovimgfilt_F,elow,ehigh)
       !
       ! produce cleaned PN OOT file if desired
       ! OLDNEW: pnS003-fovevcoot/pnS003-allevcoot
       if(withoot) then
          type = "allevcoot"
          suffix = "fits"
          call make_filename(inst,expid,type,elow,ehigh,suffix,outOOT_F,withlong)
          command = "evselect table="// trim(adjustl(inOOT_F)) // &
            " withfilteredset=yes expression='(PATTERN<=4)&&GTI(" // &
            trim(adjustl(gtifits_F)) // ",TIME)&&" // trim(adjustl(flagg)) // &
            "' filteredset="// trim(adjustl(outOOT_F)) // &
            " filtertype=expression keepfilteroutput=yes "// &
            "updateexposure=yes filterexposure=yes"
          write(errstr,*) "*Flare free ALL OOT (withoot) evselect command = "
          call message(AppMsg,Always)
          write(errstr,*) trim(adjustl(command))
          call message(AppMsg,Noisy)
          call system(trim(adjustl(command)))
          call addlohi(outOOT_F,elow,ehigh)
          call addfitpars(outOOT_F,norm,width,center,h_imme,h_totl,l_imme,l_totl)
       end if
       !
       ! create corner-only events file 
       !
       if (specialSmode)  then
          write(errstr,*) &
               "HISTOGRAM_METHOD: Create filtered corner data image and event files"
          call message(AppMsg,Verbose)
          type = "corevc"
          suffix = "fits"
          call make_filename(inst,expid,type,elow,ehigh,suffix,corevlifilt_F,withlong)
          write(errstr,*) &
               "HISTOGRAM_METHOD: *CLEANED (flare-free) CORNER EVENTLIST FILE* = "//&
               trim(adjustl(corevlifilt_F))
          call message(AppMsg,Always)
          command = "evselect table="// trim(adjustl(inEVLI_F)) //":EVENTS " // &
               "withfilteredset=yes expression='(PATTERN<="// trim(adjustl(patt)) // &
                  ")&&GTI("// trim(adjustl(gtifits_F)) // &
                  ",TIME)&&" // trim(adjustl(flagg)) // "&&" // &
                  trim(adjustl(cornerdef))//"' filteredset=" // &
                  trim(adjustl(corevlifilt_F)) // &
                  " filtertype=expression keepfilteroutput=yes " // &
                  "updateexposure=yes filterexposure=yes" 
          write(errstr,*) "*Flare free corner eventlist evselect command = "
          call message(AppMsg,Always)
          write(errstr,*) trim(adjustl(command))
          call message(AppMsg,Noisy)
          call system(trim(adjustl(command)))
          call addlohi(corevlifilt_F,elow,ehigh)
          call addfitpars(corevlifilt_F,norm,width,center,h_imme,h_totl,l_imme,l_totl)
          !
          ! create the image of the corner-only events for mos_doc.bat use
          !
          type = "corimc"
          suffix = "fits" ! GTIBUILD output FITS file
          call make_filename(inst,expid,type,elow,ehigh,suffix,corimgfilt_F,withlong)
          write(errstr,*) &
               "HISTOGRAM_METHOD: *CLEANED (flare-free) CORNER IMAGE FILE* = "//&
               trim(adjustl(corimgfilt_F))
          call message(AppMsg,Always)
          command = "evselect table="// trim(adjustl(corevlifilt_F)) //&
               " withimageset=yes imageset="// trim(adjustl(corimgfilt_F)) //&
               " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
               "ximagesize=600 yimagesize=600"
          write(errstr,*) "*Flare free corner image evselect command = "
          call message(AppMsg,Always)
          write(errstr,*) trim(adjustl(command))
          call message(AppMsg,Noisy)
          call system(trim(adjustl(command)))
       else
          write(errstr,*) &
               "HISTOGRAM_METHOD: PN window mode data so no corner "//&
               "data image or event file"
          call message(AppMsg,Verbose)
       endif
    else
       call error("noGTI","No GTI FITS file was found, HISTO method cannot continue")
    end if
    ! add residuals and fit goodness to cleaned event list
    write(errstr,*) "HISTOGRAM_METHOD: Adding residuals to cleaned event lists"
    call message(AppMsg,Always)
    write(errstr,'(a,f9.3)') "Normalization = ",norm
    call message(AppMsg,Always)
    write(errstr,'(a,f9.3)') "Width         = ",width
    call message(AppMsg,Always)
    write(errstr,'(a,f9.3)') "Center        = ",center
    call message(AppMsg,Always)
    !   
    write(errstr,*) "HISTOGRAM_METHOD: Return with no errors"
    call message(AppMsg,Always)
    return

  end subroutine histogram_method
  !
  !
  !
  subroutine ratio_method (inEVLI_F,cornerdef,anndef,elow,ehigh,obsid,&
       expid,revolut,inst,binning,smoothing,binsize,smoothfactor,&
       ratio,patt,specialSmode,rangescale,delt,fovimg_F,fovlc_F,corlc_F,&
       annevli_F,annlc_F,gtitxt_F,withlong,withoot,inOOT_F,plim)
    !
    !  added 05/2008 -- BHP from Kip Kuntz Algorithm SS input.
    !
    ! This algorithm extracts the number of counts per pixel in the 
    ! user specified (defaulted to 8.0-12.0 keV) band in an annulus 
    ! in the FOV (13900<R<16900) and compare to the counts per pixel
    ! in the corner regions for the same energy band.
    !
    ! The ratio of interest is:
    ! (annulus_counts) * corner_area / (corner_counts * annulus_area)
    !
    ! If this ratio is greater than a user specified value (defaulting
    ! to 1.2) then there is flaring.
    !
    ! Next, a set of GTIs is created in which flaring is NOT occurring,
    ! and these are used to created a filtered Event List of the fov.
    !
    type(DataSetT) :: set
    type(TableT)   :: tab
    character(len=2) :: patt, inst
    character(len=5) :: cminchan, cmaxchan
    character(len=6) :: expid, suffix
    character(len=10) :: obsid, type
    character(len=32) :: flagg
    character(len=256), intent(in) :: inEVLI_F
    character(len=256) :: fovimg_F, fovlc_F, corlc_F, annlc_F, &
         fovevlifilt_F, fovimgfilt_F, corevlifilt_F, corimgfilt_F, &
         histqdp_F, gtitxt_F, gtifits_F, annevli_F, annspec_F, &
         corevli_F, corspec_F, inOOT_F, outOOT_F
    character(len=1024) :: anndef, cornerdef, command
    integer, intent(in) :: smoothfactor, binsize
    integer :: revolut, method, elow, ehigh
    logical :: specialSmode, withoot
    logical, intent(in) :: binning, smoothing, withlong
    logical :: f_ok = .false.
    real(kind=double) :: cor_area, ann_area, cor_cnt, &
         ann_cnt, newratio, cor_time, ann_time, unc, &
         h_imme, h_totl, l_imme, l_totl
    real(kind=double) :: rangescale, delt
    real(kind=single),intent(in) :: ratio
    real(kind=double) :: norm, width, center, apar(3), plim(2)
    !
    method = 2 
    !    rangescale = 1.0 ! still hardcoded for ratio method???!!!
    if (binsize .lt. 100) then
       call warning("RatioBinLT100",&
            "Binsize for Ratio Method should be >= 100")
    end if

    if (.not.specialSmode) then
       write(errstr,*) &
            "RATIO_METHOD: PN window mode - try the histogram method"
       call message(AppMsg,Sparse)
    endif
    !
    write(errstr,*) "RATIO_METHOD: Chip corner definition:  ",trim(adjustl(cornerdef))
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: Chip annulus definition: ",trim(adjustl(anndef))
    call message(AppMsg,Sparse)
    !
    ! kk does the below command, but 8000 and 12000 are now user-selected. 
    ! he also takes an already-filtered-for-cornerdef eventfile, whereas
    ! it now takes from a raw event list and does both cornerdef and others.
    ! command = "evselect table=mos".$ins_num."-corn.fits:EVENTS ".
    ! "withfilteredset=yes expression='(PI in [8000:12000])'".//&
    !  filtertype=expression "   
    ! command = "evselect table=" // trim(corevfileF) // ":EVENTS " //&
    ! if use evfile, add cornerdef?
    ! so, raw eventlist, filter on energy
    !
    write(errstr,*) "RATIO_METHOD: extract corner spectrum with evselect"
    call message(AppMsg,Always)
    type = "corev"  ! mos1S001-corev.fits
    suffix = "fits" ! corner unfiltered eventlist created at beginning
    call make_filename(inst,expid,type,elow,ehigh,suffix,corevli_F,withlong)
    type = "cor"  ! mos1S001-cor.pi
    suffix = "pi" ! corner spectrum
    call make_filename(inst,expid,type,elow,ehigh,suffix,corspec_F,withlong)
    write(errstr,*) "RATIO_METHOD: *CORNER SPECTRUM FILE* = "//&
         trim(adjustl(corspec_F))
    call message(AppMsg,Always)
    !
    command = "evselect table=" // trim(adjustl(corevli_F)) // &
         " withfilteredset=yes expression='TRUE'"//&
         " filtertype=expression "// &
         "keepfilteroutput=no updateexposure=yes filterexposure=yes "//&
         "withspectrumset=yes spectrumset=" // trim(adjustl(corspec_F)) //&
         " energycolumn=PI spectralbinsize=15 withspecranges=yes "//&
         "specchannelmin=0 specchannelmax=11999"
    !
    ! should specchans be set by user? Aren't they redundant? Yes; The
    ! event list was created with cmin/cmaxchan so no need to even
    ! have the above...
    !
    write(errstr,*) "*Corner spectrum evselect command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(adjustl(command)))
    !
    ! command = "backscale spectrumset=mos".$ins_num.
    ! "-corn.pi badpixlocation=mos".$ins_num.
    ! "-clean.fits withbadpixcorr=yes ignoreoutoffov=no";
    !
    ! backscale the corner spectrum created above to get the BACKSCAL
    ! keyword (the area of the corners used to make the corner spectrum.
    ! as from the task Backscale: area = geometric_area - ccd_gaps - bad_pixels)
    !
    ! run backscale on the corner spectrum (3)
    !
    write(errstr,*) "RATIO_METHOD: Backscale the corner spectrum."
    call message(AppMsg,Always)
    command = "backscale spectrumset=" // trim(adjustl(corspec_F)) //&
         " badpixlocation=" // trim(adjustl(inEVLI_F)) //&
         " withbadpixcorr=yes ignoreoutoffov=no";
    write(errstr,*) "*Corner spectrum backscale command = "
    call message(AppMsg,Noisy)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Noisy)
    call system(trim(adjustl(command)))
    !
    ! ***FOV annulus event list creation moved to makeevimlc subroutine!***
    !
    ! next, make the FOV annulus spectrum (5)
    !
    ! already applied pattern, flag and cornerdef, so just energy,
    ! binsize, etc needed:
    !
    ! OLDNEW: fovannspecF = mos1S001-annspec.FIT/mos1S001-ann.pi
    !
    type = "ann"
    suffix = "pi" ! FOV annulus spectrum
    call make_filename(inst,expid,type,elow,ehigh,suffix,annspec_F,withlong)
    write(errstr,*) "RATIO_METHOD: FOV annulus spectral file =",&
         trim(adjustl(annspec_F))
    call message(AppMsg,Always)
    write(errstr,*) "RATIO_METHOD: Get Flare discriminator information"
    call message(AppMsg,Noisy)
    write(errstr,*) "RATIO_METHOD: Evselect extract FOV annulus spectrum"
    call message(AppMsg,Noisy)
    !
    command = "evselect table=" // trim(adjustl(annevli_F)) // &
         " withfilteredset=yes expression='TRUE'"//&
         " filtertype=expression " // &
         "keepfilteroutput=no updateexposure=yes filterexposure=yes "//&
         "withspectrumset=yes spectrumset=" // trim(adjustl(annspec_F)) //&
         " energycolumn=PI spectralbinsize=15 withspecranges=yes "//&
         "specchannelmin=0 specchannelmax=11999"
    write(errstr,*) "*FOV annulus spectrum evselect command = "
    call message(AppMsg,Noisy)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Noisy)
    call system(trim(adjustl(command)))
    !
    ! $comm="evselect table=mos".$ins_num."-clean.fits:EVENTS ".
    ! I don't get this 'clean' file. sounds like an espfilt file
    ! cleaned but doesn't evselect do the right thing below on the
    ! raw event file?
    !
    ! BHP - use *raw* emchain output eventlist, fovanndef cleans it...
    !
    !kk get the flare discriminator information
    !kk extract the FOV annulus spectrum
    !
    ! original:
    ! command="backscale spectrumset=mos".$ins_num."-fovann.pi "//&
    !         "badpixlocation=mos".$ins_num."-clean.fits withbadpixcorr=yes "//&
    !         "ignoreoutoffov=no";
    !
    ! BACKSCAL = area of region used to make spectral file 
    ! now run backscale on fov annulus spectrum (6)
    !
    write(errstr,*) "RATIO_METHOD: Running Backscale on FOV Annulus spectrum."
    call message(AppMsg,Always)
    !
    ! fovspec = P0099840101M1S001-annspec.FIT 
    !
    command="backscale spectrumset=" // trim(adjustl(annspec_F)) // &
         " badpixlocation=" // trim(adjustl(annevli_F)) // &
         " withbadpixcorr=yes ignoreoutoffov=no"
    write(errstr,*) "*FOV annulus spectrum backscale command = "
    call message(AppMsg,Noisy)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Noisy)
    call system (trim(adjustl(command)))
    !
    ! kk get the 8-12keV flux and backscale-the new way
    !  bp get it via the OAL
    ! kk command = "fkeypar " // trim(corspecF) 
    ! kk get some header keywords to create output file's names (prefix)
    !  bp clean/check here (look for hasblock, etc) 
    !
    set = dataSet(corspec_F, READ)
    tab = table(set,"SPECTRUM")
    cor_area = real32Attribute(tab,"BACKSCAL") !corn_scl
    call release(set)
    write(errstr,*) "RATIO_METHOD: Corner Area:",cor_area
    call message(AppMsg,Sparse)
    !
    set = dataSet(annspec_F, READ)
    tab = table(set,"SPECTRUM")
    ann_area = real32Attribute(tab,"BACKSCAL") !annu_scl
    call release(set)
    write(errstr,*) "RATIO_METHOD: FOV annulus Area:",ann_area
    call message(AppMsg,Sparse)
    !
    set = dataSet(corevli_F, READ)
    tab = table(set,"EVENTS",READ)
    cor_cnt = real(numberOfRows(tab))
    call release(set)
    write(errstr,*) "RATIO_METHOD: Corner total counts:",cor_cnt
    call message(AppMsg,Sparse)
    !
    set = dataSet(annevli_F, READ)
    tab = table(set,"EVENTS",READ)
    ann_cnt = real(numberOfRows(tab))
    call release(set)
    write(errstr,*) "RATIO_METHOD: FOV annulus total counts:",ann_cnt
    call message(AppMsg,Sparse)
    !
    set = dataSet(corspec_F, READ)
    tab = table(set,"SPECTRUM")
    cor_time = real32Attribute(tab,"EXPOSURE") !time_cor
    call release(set)
    write(errstr,*) "RATIO_METHOD: Corner total exposure:",cor_time
    call message(AppMsg,Sparse)
    !
    set = dataSet(annspec_F, READ)
    tab = table(set,"SPECTRUM")
    ann_time = real32Attribute(tab,"EXPOSURE") !time_ann
    call release(set)
    write(errstr,*) "RATIO_METHOD: FOV annulus total exposure:",ann_time
    call message(AppMsg,Sparse)

    ! After you create the event files for each of the regions,
    ! 1. extract a spectrum, and "backscale" it (run backscale). 
    ! 2. from the header of the spectrum:
    !    time = value in exposure keyword
    !    area = value in backscal keyword
    ! Watch out for the area as this is a large number (essentially the number 
    ! of pixels) so you will need long or something.
    !
    !
    ! What I have done is extract the number of counts per pixel in 8.0-12.0 keV
    ! in an annulus in the FOV (13900<R<16900) and compare to the counts per pixel
    ! in the corner regions for the same energy band.
    ! The ratio of interest is $annu_cnt * $corn_scl/($corn_cnt * $ annu_scl)
    ! If this ratio is greater than 1.2 then you are definitely flared.
    !
    write(errstr,*) "RATIO_METHOD: Calculating ratio:"
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: ratio = (fovann_cnts/corner_counts) * "
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: (corner_time*corner_area/corner_counts) / "
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: (fovann_time*fovann_area)"
    call message(AppMsg,Sparse)
    newratio = ann_cnt/cor_cnt * &
         (cor_time*cor_area)/(ann_time*ann_area)
    unc = sqrt(((ann_cnt/(cor_cnt**2)+(1/cor_cnt**3))) * &
         ((cor_time*cor_area)/(ann_time*ann_area)))
    !
    write(errstr,*) "RATIO_METHOD: user-selected or default ratio = ",ratio
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: data-derived ann/cor ratio     = ",newratio
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: uncertainty in newratio        = ",unc
    call message(AppMsg,Sparse)
    !
    ! call PGPLOT directly in the future? Eliminate QDP file construct!!!
    !
    ! OLDNEW: histqdpF = mos1S001-hist.qdp" ! QDP output file
    !type = "hist"
    !suffix = "qdp" 
    !call make_filename(inst,expid,type,elow,ehigh,suffix,histqdp_F,withlong)
    !write(errstr,*) "RATIO_METHOD: *HISTOGRAM QDP OUTPUT FILE* = "//&
    !     trim(adjustl(histqdp_F))
    !call message(AppMsg,Always)
    !gtitxtF = mos1S001-gti.txt" ! GTI temp text file
    type = "gti"
    suffix = "txt"
    call make_filename(inst,expid,type,elow,ehigh,suffix,gtitxt_F,withlong)
    write(errstr,*) "RATIO_METHOD: *GTI TEXT FILE* = "//&
         trim(adjustl(gtitxt_F))
    call message(AppMsg,Noisy)
    ! gtifitsF = mos1S001-gti.FIT"! GTIBUILD output FITS file
    type = "gti"
    suffix = "fits" 
    call make_filename(inst,expid,type,elow,ehigh,suffix,gtifits_F,withlong)
    write(errstr,*) "RATIO_METHOD: *GTI FITS FILE* = "//&
         trim(adjustl(gtifits_F))
    call message(AppMsg,Always)
    !
    ! clean_lc is the engine that creates the gti txt file for use by
    ! gtibuild and the smoothed, binned lc's
    !
    method = 2 ! histogram = 1, ratio = 2 !!! fix this
    write(errstr,*) "RATIO_METHOD: Running Clean_LC..."
    call message(AppMsg,Verbose)
    call clean_lc(obsid,expid,gtitxt_F,fovlc_F,corlc_F,annlc_F,histqdp_F,&
         binning,smoothing,binsize,smoothfactor,method,newratio,specialSmode,&
         rangescale,delt,norm,width,center,apar,plim,h_imme,h_totl,l_imme,l_totl)

    write(errstr,*) "RATIO_METHOD: *Running GTIBUILD* = "
    call message(AppMsg,Noisy)
    command = "gtibuild file="// trim(adjustl(gtitxt_F)) //&
         " table="// trim(adjustl(gtifits_F))
    call system(trim(adjustl(command)))
    !
    ! if GTI file exists, create filtered, flare-free event list
    !
    if (inst .eq. 'PN') then ! new for SAS-19
       flagg =  '(#XMMEA_EP)'
    else
       flagg = "((FLAG & 0x766aa000)==0)"
    endif
    inquire(file=trim(adjustl(gtifits_F)),exist=f_ok)
    if (f_ok) then
       ! OLDNEW: fovevlifiltF = mos1S001-fovevlifilt.FIT/mos1S001-allevc.fits was: fovevc
       type = "allevc"
       suffix = "fits" 
       call make_filename(inst,expid,type,elow,ehigh,suffix,fovevlifilt_F,withlong)
       write(errstr,*) "RATIO_METHOD: *FILTERED ALL EVENT LIST FILE* = "//&
            trim(adjustl(fovevlifilt_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(inEVLI_F)) // &
            " withfilteredset=yes "// &
            "expression='(PATTERN<=" // trim(adjustl(patt)) // ")&&GTI(" // &
            trim(adjustl(gtifits_F)) // ",TIME)&&" // trim(adjustl(flagg)) // &
            "' filteredset="// trim(adjustl(fovevlifilt_F)) // &
            " filtertype=expression keepfilteroutput=yes "// &
            "updateexposure=yes filterexposure=yes"
       call system(trim(adjustl(command)))
       call addlohi(fovevlifilt_F,elow,ehigh)
       call addfitpars(fovevlifilt_F,norm,width,center,h_imme,h_totl,l_imme,l_totl)
       ! OLDNEW: mos1S001-fovimgfilt.FIT/mos1S001-allimc.fits was: fovimc
       type = "allimc"
       suffix = "fits" 
       call make_filename(inst,expid,type,elow,ehigh,suffix,fovimgfilt_F,withlong)
       write(errstr,*) "RATIO_METHOD: *FILTERED ALL IMAGE FILE* = "//&
            trim(adjustl(fovimgfilt_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(fovevlifilt_F)) //&
            " withimageset=yes imageset="// trim(adjustl(fovimgfilt_F)) //&
            " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
            "ximagesize=600 yimagesize=600"
       call system(trim(adjustl(command)))
       !
       ! create cleaned corner-only events file 
       !
       ! OLDNEW: corevlifiltF = mos1S001-corevlifilt.FIT/mos1S001-corevc.fits
       !
       type = "corevc"
       suffix = "fits" 
       call make_filename(inst,expid,type,elow,ehigh,suffix,corevlifilt_F,withlong)
       write(errstr,*) "RATIO_METHOD: *FILTERED CORNER EVENT LIST FILE* = "//&
            trim(adjustl(corevlifilt_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(inEVLI_F)) //":EVENTS " // &
            "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(patt)) // &
            ")&&GTI(" // trim(adjustl(gtifits_F)) // ",TIME)&&" // &
            trim(adjustl(flagg)) // "&&" // trim(adjustl(cornerdef)) // &
            "' filteredset="// trim(adjustl(corevlifilt_F)) // &
            " filtertype=expression keepfilteroutput=yes " // &
            "updateexposure=yes filterexposure=yes"
       write(errstr,*) "*Flare free corner eventlist evselect command = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(corevlifilt_F,elow,ehigh)
       call addfitpars(corevlifilt_F,norm,width,center,h_imme,h_totl,l_imme,l_totl)
       ! create the image of the corner-only events for mos_doc.bat use
       ! OLDNEW: corimgfiltF = mos1S001-corimgfilt.FIT/mos1S001=corimc
       type = "corimc"
       suffix = "fits" 
       call make_filename(inst,expid,type,elow,ehigh,suffix,corimgfilt_F,withlong)
       write(errstr,*) "RATIO_METHOD: *FILTERED CORNER IMAGE FILE* = "//&
            trim(adjustl(corimgfilt_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(corevlifilt_F)) //&
            " withimageset=yes imageset="// trim(adjustl(corimgfilt_F)) //&
            " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
            "ximagesize=600 yimagesize=600"
       write(errstr,*) "*Flare free corner image evselect command = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
    else
       call error("noGTI","No GTI FITS file was found, cannot continue")
    end if
    !
    ! produce cleaned PN OOT file if desired
    !
    if(withoot) then
       outOOT_F = "pn" // trim(adjustl(expid)) // "allevcoot.fits" 
       command = "evselect table="// trim(adjustl(inOOT_F)) // &
            " withfilteredset=yes "// &
            "expression='(PATTERN<=4)&&GTI(" // &
            trim(adjustl(gtifits_F)) // ",TIME)&&" // trim(adjustl(flagg)) // &
            "' filteredset="// trim(adjustl(outOOT_F)) // &
            " filtertype=expression keepfilteroutput=yes "// &
            "updateexposure=yes filterexposure=yes"
       write(errstr,*) "*Flare free ALL OOT evselect command = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(outOOT_F,elow,ehigh)
       call addfitpars(outOOT_F,norm,width,center,h_imme,h_totl,l_imme,l_totl)
    end if

    write(errstr,*) "RATIO_METHOD: Return with no errors"
    call message(AppMsg,Always)
    return

  end subroutine ratio_method
  !
  !
  !
  subroutine clean_lc(obsid,expid,gtitxtFile,fovlcFile,corlcFile,annlcFile,&
       histqdpFile,binning,smoothing,binsize,smoothfactor,method,ratio,&
       specialSmode,rangescale,delt,norm,width,center,apar,plim,h_imme,&
       h_totl,l_imme,l_totl)
    
    type(TableT)  :: fovlc_T, corlc_T, annlc_T
    type(ColumnT) :: time_lcCol, rate_lcCol
    type(ColumnT) :: time_clcCol, rate_clcCol
    type(ColumnT) :: time_flcCol, rate_flcCol
    type(DataSetT) :: fovlc_S, corlc_S, annlc_S
    character(len=6) :: expid
    character(len=8) :: cdif1, cdif2, cdif3, cdif4
    character(len=10) :: obsid
    character(len=256) :: fovlcFile, corlcFile, annlcFile, &
         histqdpFile, gtitxtFile
    integer :: kip_rest,kip_ierr
    real(kind=double) :: kip_parms(3)
    integer, intent(in) :: binsize, smoothfactor, method
    integer :: mask(3), binwidth, smoo, nverls, nbins, &
         delta, low, high
    integer, parameter :: ndim = 3
    integer :: i, ii, it, j, k, nn, np, imaxp, n_hist
    integer :: nbins_lc, nbins_clc, nchan_lc, nchan_clc, &
         nbins_flc, nchan_flc, binned_lc, binned_clc, binned_flc
    integer, allocatable, dimension(:) :: maxi, ihist_cnts, plc, plc2, itemp
    logical, intent(in) :: binning, smoothing, specialSmode
    logical :: badl, badh
    real(kind=double) :: rangescale, delt
    real(kind=double), intent(in) :: ratio
    real(kind=double) :: maxp, maxr, th, rlow, rhigh, parms(3), &
         dparms(3), v2, chilo, chihi    
    real(kind=double) :: currchan, mintime
    real(kind=double), allocatable, dimension(:) :: hist_cnts, hist_rate, &
         gauss_ind, gauss_dep, sortindex, smoo_time, smoo_tim2, temp, &
         v1, hist_modl, hist_resi, tempsortindex
    real(kind=double) :: apar(3), norm, width, center, winw, bin0, &
         binw, plim(2)
         ! was hardcoded here pre-19 plim(3) = (/0.1d0, 5.0d0, 1.0d0/)
    real(kind=double) :: reslo(3), treslo(3), reshi(3), h_totl, l_totl, h_imme, l_imme
    real(kind=double) :: kip_reslo(3), kip_chilo, tchilo
    real(kind=double), allocatable, dimension(:) :: &
         tstart_lc, tend_lc, tstart_clc, tend_clc, &
         tstart_flc, tend_flc, tempstart, tempend
    real(kind=double), allocatable, dimension(:) :: &
         new_modrate_lc, new_modrate_clc, new_modrate_flc
    real(kind=double), allocatable, dimension(:) :: &
         modrate_lc, modrate_clc, modrate_flc
    real(kind=double), allocatable, dimension(:) :: &
         time_lc, time_clc, time_flc, rate_lc, rate_clc, rate_flc
    real(kind=double), dimension(:), pointer   :: &
         p_time_lc, p_rate_lc, p_time_clc, p_rate_clc,&
         p_time_flc, p_rate_flc

    apar    = 0.0
    it    = 0
    rlow  = 0.
    rhigh = 0.
    chilo = 0.d0
    chihi = 0.d0
    norm   = 0.d0
    width  = 0.d0
    center = 0.d0
    mask = (/1,1,1/) ! tells amoeba which variables to fit
    !
    ! set up binning and smoothing 
    !
    if (.not.binning) then
       binwidth = 1
       write(errstr,*) "CLEAN_LC: Binning factor of 1."
       call message(AppMsg,Verbose)    
    else 
       binwidth = binsize
    end if
    if (.not.smoothing) then
       smoo = 1
       write(errstr,*) "CLEAN_LC: Smoothing factor of 1."
       call message(AppMsg,Verbose)    
    else 
       smoo = smoothfactor
    end if
    !!! AI: Fix to avoid Out of Memory error in some 64 bit platforms
    nchan_flc = 32767 ! what if binning = 2 or something???
    !   
    ! open and read the FOV lightcurve, extracting TIME and RATE only
    !
    fovlc_S = dataSet(trim(adjustl(fovlcFile)), READ) ! mos1S001-fovlc.fits
    if (.not.hasBlock(fovlc_S,"RATE")) then
       call error("badInfile","FOV LC does not contain a RATE extension")
    else
       fovlc_T = table(fovlc_S,"RATE")
    endif
    nchan_lc = numberOfrows(fovlc_T)
    write(errstr,*) "CLEAN_LC: There are ",nchan_lc," rows in the FOV LC"
    call message(AppMsg,Always)
    allocate(time_lc(nchan_lc),rate_lc(nchan_lc))
    time_lc = 0.
    rate_lc = 0.
    if (.not.hasColumn(fovlc_T,"TIME")) then
       call error("badTable","RATE table does not contain a TIME column")
    else
       time_lcCol = column(fovlc_T,"TIME",READ)
       p_time_lc => Real64Data(time_lcCol)
       time_lc(1:nchan_lc) = p_time_lc(0:nchan_lc-1)
    endif
    nullify (p_time_lc)
    if (.not.hasColumn(fovlc_T,"RATE")) then
       call error("badTable","RATE table does not contain a RATE column")
    else
       rate_lcCol = column(fovlc_T,"RATE",READ)
       p_rate_lc => Real64Data(rate_lcCol)
       rate_lc(1:nchan_lc) = p_rate_lc(0:nchan_lc-1)
    endif
    nullify (p_rate_lc)
    call release(rate_lcCol)
    call release(fovlc_T)
    call release(fovlc_S)
    !
    ! open and read the corner lightcurve, extracting TIME and RATE only
    !
    if(specialSmode) then ! not EPN/PFW/FFExtended
       corlc_S = dataSet(trim(adjustl(corlcFile)), READ) 
       if (.not.hasBlock(corlc_S,"RATE")) then
          call error("badInfile",&
               "Corner LC does not contain an RATE extension")
       else
          corlc_T = table(corlc_S,"RATE")
       endif
       nchan_clc = numberOfrows(corlc_T)
       write(errstr,*) "CLEAN_LC: There are ",nchan_clc,&
            " rows in the Corner LC."
       call message(AppMsg,Verbose)
       allocate(time_clc(nchan_clc),rate_clc(nchan_clc))
       time_clc = 0.
       rate_clc = 0.
       if (.not.hasColumn(corlc_T,"TIME")) then
          call error("badTable","RATE table does not contain a TIME column")
       else
          time_clcCol = column(corlc_T,"TIME",READ)
          p_time_clc => Real64Data(time_clcCol)
          time_clc(1:nchan_clc) = p_time_clc(0:nchan_clc-1)
       endif
       nullify (p_time_clc)
       if (.not.hasColumn(corlc_T,"RATE")) then
          call error("badTable","RATE table does not contain a RATE column")
       else
          rate_clcCol = column(corlc_T,"RATE",READ)
          p_rate_clc => Real64Data(rate_clcCol)
          rate_clc(1:nchan_clc) = p_rate_clc(0:nchan_clc-1)
       endif
       nullify (p_rate_clc)
       call release(time_clcCol)
       call release(rate_clcCol)
       call release(corlc_T)
       call release(corlc_S)
    else ! dummy if not specialSmode
       allocate(time_clc(1),rate_clc(1))
       allocate(modrate_clc(1),new_modrate_clc(1))
       allocate(tstart_clc(1),tend_clc(1))
       time_clc = 0.
       rate_clc = 0.
       modrate_clc = 0.
       new_modrate_clc = 0.
       tstart_clc = 0.
       tend_clc = 0.
    endif
    !
    ! do the below IF block for RATIO METHOD only
    ! open and read FOV annulus lightcurve, extracting TIME and RATE only
    !
    binned_flc = 0
    if (method .eq. 2) then
       annlc_S = dataSet(trim(adjustl(annlcFile)), READ)
       if (.not.hasBlock(annlc_S,"RATE")) then
          call error("badInfile",&
               "FOV annulus LC does not contain an RATE extension")
       else
          annlc_T = table(annlc_S,"RATE")
       endif
       nchan_flc = numberOfrows(annlc_T)
       write(errstr,*) "CLEAN_LC: There are ",nchan_flc,&
            " rows in the FOV annulus LC."
       allocate (time_flc(nchan_flc),rate_flc(nchan_flc))
       time_flc = 0.
       rate_flc = 0.
       if (.not.hasColumn(annlc_T,"TIME")) then
          call error("badTable","RATE table does not contain a TIME column")
       else
          time_flcCol = column(annlc_T,"TIME",READ)
          p_time_flc => Real64Data(time_flcCol)
          time_flc(1:nchan_flc) = p_time_flc(0:nchan_flc-1)
       endif
       nullify (p_time_flc)
       if (.not.hasColumn(annlc_T,"RATE")) then
          call error("badTable","RATE table does not contain a RATE column")
       else
          rate_flcCol = column(annlc_T,"RATE",READ)
          p_rate_flc => Real64Data(rate_flcCol)
       endif
       nullify (p_rate_flc)
       call release(time_flcCol)
       call release(rate_flcCol)
       call message(AppMsg,Verbose)
       call release(annlc_T)
       call release(annlc_S)
       !
       allocate(modrate_flc(nchan_flc),new_modrate_flc(nchan_flc))
       allocate(tempstart(nchan_flc),tempend(nchan_flc))
       modrate_flc = 0.
       new_modrate_flc = 0.
       tempstart = 0.
       tempend = 0.
       write(errstr,*) &
            "CLEAN_LC: Calling binit to bin up the FOV annulus LC."
       call message(AppMsg,Verbose)
       call binit(time_flc,rate_flc,binwidth,nchan_flc, &
            modrate_flc,tempstart,tempstart,binned_flc)
       allocate(tstart_flc(binned_flc),tend_flc(binned_flc))
       tstart_flc = 0.
       tend_flc = 0.
       tstart_flc(1:binned_flc) = tempstart(1:binned_flc)
       tend_flc(1:binned_flc) = tempend(1:binned_flc)
       deallocate(tempstart,tempend)
       nchan_flc = nchan_flc -1
       binned_flc = binned_flc -1
       write(errstr,*) "CLEAN_LC: Smoothing the FOV annulus LC"
       call message(AppMsg,Noisy)
       call smooth(rate_flc,nchan_flc,smoo,new_modrate_flc)
       deallocate(rate_flc)
    else ! dummies for histogram method. 
       allocate(time_flc(1),rate_flc(1))
       allocate(modrate_flc(1),new_modrate_flc(1))
       allocate(tstart_flc(1),tend_flc(1))
       time_flc = 0.
       rate_flc = 0.
       modrate_flc = 0.
       new_modrate_flc = 0.
       tstart_flc = 0.
       tend_flc = 0.
    end if
    !
    write(errstr,*) "CLEAN_LC: Allocating arrays for FOV LC."
    call message(AppMsg,Verbose)
    !
    ! modrate is used for plotting, new_modrate is the smoothed
    !
    binned_lc = 0
    allocate(modrate_lc(nchan_lc),new_modrate_lc(nchan_lc))
    allocate(tempstart(nchan_lc),tempend(nchan_lc))
    modrate_lc = 0.d0
    new_modrate_lc = 0.d0
    tempstart = 0.d0
    tempend = 0.d0
    call message(AppMsg,Verbose)
    write(errstr,*) "CLEAN_LC: Calling binit to bin up the FOV LC."
    call message(AppMsg,Verbose)
    ! binit is in goflib/goflib_utils.f90
    call binit(time_lc,rate_lc,binwidth,nchan_lc,modrate_lc, &
         tempstart,tempend,binned_lc)
    do i=1,nchan_lc
       write(errstr,*)time_lc(i),rate_lc(i),modrate_lc(i),tempstart(i),tempend(i)
       call message(AppMsg,Noisy)
    end do
    allocate(tstart_lc(binned_lc),tend_lc(binned_lc))
    tstart_lc = 0.
    tend_lc = 0.
    tstart_lc(1:binned_lc) = tempstart(1:binned_lc)
    tend_lc(1:binned_lc) = tempend(1:binned_lc)
    deallocate(tempstart,tempend)
    nchan_lc = nchan_lc - 1       
    binned_lc = binned_lc - 1
    !
    if(specialSmode) then
       binned_clc = 0
       allocate(modrate_clc(nchan_clc),new_modrate_clc(nchan_clc))
       allocate(tempstart(nchan_clc),tempend(nchan_clc))
       call message(AppMsg,Verbose)
       write(errstr,*) "CLEAN_LC: Calling binit to bin up the FOV LC."
       call message(AppMsg,Verbose)
       call binit(time_clc,rate_clc,binwidth,nchan_clc,modrate_clc, &
            tempstart,tempend,binned_clc)
       allocate(tstart_clc(binned_clc),tend_clc(binned_clc))
       tstart_clc(1:binned_clc) = tempstart(1:binned_clc)
       tend_clc(1:binned_clc) = tempend(1:binned_clc)
       deallocate(tempstart,tempend)
       nchan_clc = nchan_clc - 1       
       binned_clc = binned_clc - 1
    end if
    !
    ! Smooth the data for fitting the distribution
    ! smoo = 1 if not.smoothing 
    !
    write(errstr,*) "CLEAN_LC: Smoothing the FOV LC"
    call message(AppMsg,Noisy)
    write(errstr,*)'XXXXX size of rate_lc = ',size(rate_lc)
    call message(AppMsg,Always)
    write(errstr,*)'XXXXX nchan_lc = ',nchan_lc
    call message(AppMsg,Always)
    write(errstr,*)'XXXXX size of new_modrate_lc = ',size(new_modrate_lc)
    call message(AppMsg,Always)
    call smooth(rate_lc,nchan_lc,smoo,new_modrate_lc)
    deallocate(rate_lc)
    !
    if(specialSmode) then
       write(errstr,*) "CLEAN_LC: Smoothing the Corner LC"
       call message(AppMsg,Noisy)
       call smooth(rate_clc,nchan_clc,smoo,new_modrate_clc)
    endif
    deallocate(rate_clc)
    !
    ! the below is essentially a rewrite of IDL clean_rel.pro (BP, 2020)
    !
    binw = 1.0d0 / 51.d0       ! was 1/60, should be 1.0/smoo (KDK)
    bin0 = plim(1)
    nbins = int(plim(2)*2/ binw) !KDK
    write(errstr,*)"Starting Multidimensional minimzation of fit_gauss (amoeba)"
    call message(AppMsg,Always)
1010 format(a,i5,a,f9.4)
1012 format(a,3f9.4,a,f9.4)
    write(errstr,1012) 'plim(1:2) =',plim(1),plim(2)
    call message(AppMsg,Always)
    write(errstr,1010)&
         '**** sortarray max bins    = ',nbins,' histo binwidth = ',binw
    call message(AppMsg,Always)
    allocate(tempsortindex(nbins))
    nverls = 0 ! actual number of histo intervals
    !
    ! fill sorting index by intervals
    !
    do i=0,nbins-1
       currchan = plim(1) + i * binw
       if (currchan .gt. plim(2)*2) then !KDK
          exit
       else
          tempsortindex(i+1) = currchan
       endif
       nverls = nverls + 1
    end do

    if (nverls .lt. 1) then 
       call warning ("noIntervals","Did not find any intervals that meet criteria")
    end if
    allocate(sortindex(nverls))
    sortindex(1:nverls) = tempsortindex(1:nverls)
    deallocate(tempsortindex)
    write(errstr,'(a,i5)')'**** sortarray actual bins (nverls) = ',nverls
    call message(AppMsg,Always)
    !
    ! build the histogram
    ! nverls + 1 because histo 'bucket' needs that
    allocate(itemp(nverls+1),hist_cnts(nverls),hist_rate(nverls))
    allocate(maxi(nverls))
    itemp = 0
    hist_cnts  = 0. 
    hist_rate  = 0.
    write(errstr,'(4(a,f8.3))')"min LC value= ",minval(new_modrate_lc),&
         ' max LC value = ',maxval(new_modrate_lc),' limits(2) = ',plim(2)
    call message(AppMsg,Always)
    !
    ! subroutine histo is in goflib/gof_utilities_mod.f90
    !
    call histo(new_modrate_lc,nchan_lc,sortindex,nverls,itemp)
    allocate(ihist_cnts(nverls-1))
    ihist_cnts(1:nverls) = itemp(1:nverls)
    ihist_cnts(1)=0
    deallocate(sortindex,itemp) ! no longer needed
    do j=0,nverls-1 ! replaces findgen call from clean_rel.pro
       hist_rate(j+1) = j * binw + plim(1)
       if (ihist_cnts(j+1) .gt. 0) then 
          write(errstr,1987) 'hist_cnts(',j+1,') = ',ihist_cnts(j+1),&
               ' hist_rate(',j+1,') = ',hist_rate(j+1)       
          call message(AppMsg,Verbose)
       end if
    enddo
1987 format(a,i3,a,i5,a,i3,a,f8.3)
    if (sum(ihist_cnts) .eq. 0) then 
       call error ("noCounts",&
            "All histo counts are zero! Check your FOV Lightcurve!") ! improve this!!!???
    end if
    maxi   = maxloc(ihist_cnts)  ! array location of peak
    maxp   = real(maxi(1),8)
    imaxp  = maxi(1)
    maxr   = hist_rate(imaxp)    ! rate at peak location
    rlow   = max(1.,maxp-delt+1) ! delt was hardcoded to 12, now a parameter
    rhigh  = min(maxp+delt+1,real(nverls)) 
    low    = int(rlow)
    high   = int(rhigh) 
    delta  = high - low + 1
    write(errstr,1916) &
         '***** pre-amoeba: low =',low,' peak = ',imaxp,' high = ',high,' delta = ',delta
1916 format(a,i5,a,i5,a,i5,a,i5)
    call message(AppMsg,Always)
    write(errstr,1916)'***** pre-amoeba: cnts(low) =',ihist_cnts(low),&
         ' cnts(peak) = ',ihist_cnts(imaxp),' cnts(high) = ',ihist_cnts(high)
    call message(AppMsg,Always)
    allocate(gauss_ind(delta),gauss_dep(delta))
    !
    kip_n=delta
    allocate(kip_x(delta),kip_y(delta))
    !
    j = 1
    do i=low,high
       gauss_ind(j) = real(hist_rate(i),8)   ! nint???
       gauss_dep(j) = real(ihist_cnts(i),8)  ! nint???
       kip_x(j)=real(hist_rate(i),8)
       kip_y(j)=real(ihist_cnts(i),8)
       write(errstr,'(a,i3,a,g10.4,a,i3,a,g10.4)')"pre-amoeba gauss_ind(",j,") = ",gauss_ind(j),&
            "gauss_dep(",j,") = ",gauss_dep(j)
       call message(AppMsg,Verbose)
       j = j + 1
    end do
    parms = (/ real(ihist_cnts(imaxp),8), real(hist_rate(imaxp),8), 0.1D0 /)
    dparms = parms / 10.d0
    write(errstr,'(a,3f9.3)')" pre-amoeba parms = ",parms
    call message(AppMsg,Always)    
1919 format(a,i5,a,i5,a,i5)
    !
    ! first iteration of kpamoeba 111111111111
    !
    write(errstr,*)'ZZZ Pre amoeba'
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st low       = ',low
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st high      = ',high
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st ndim      = ',ndim
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st parms     = ',parms
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st dparms    = ',dparms
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st mask      = ',mask
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st reslo     = ',reslo
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st reshi     = ',reshi
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st it        = ',it
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st chilo     = ',chilo
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st chihi     = ',chihi
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st gaussind  = ',gauss_ind
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st gaussdep  = ',gauss_dep
    call message(AppMsg,Noisy)
    write(errstr,*)'pre1st delta     = ',delta
    call message(AppMsg,Noisy)
    write(errstr,*)''
    call message(AppMsg,Noisy)

    kip_parms=parms
    !call kpamoeba(ndim,parms,dparms,mask,0.01d0,100,reslo,reshi,it,chilo,&
    !     chihi,gauss_ind,gauss_dep,delta)

    !print *,"kip old1: ",reslo,chilo
    !treslo = reslo
    !tchilo = chilo
    call nelmin(kip_gauss,ndim,kip_parms,reslo,chilo,0.010d0,dparms,10,100,it,kip_rest,kip_ierr)
    !print *,"kip new1: ",reslo,chilo

    !call diffit(treslo(1),reslo(1),cdif1)
    !call diffit(treslo(2),reslo(2),cdif2)
    !call diffit(treslo(3),reslo(3),cdif3)
    !call diffit(tchilo,chilo,cdif4)

    write(errstr,*)"1st pass:"
    call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)
    !write(errstr,1882)"old         : ",treslo(1),treslo(2),treslo(3),tchilo
    !call message(AppMsg,Always)
    write(errstr,1882)"new         : ",reslo(1),reslo(2),reslo(3),chilo
    call message(AppMsg,Always)
    !write(errstr,1883)"percent diff: ",cdif1,"%",cdif2,"%",cdif3,"%",cdif4,"%"
    !call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)
1882 format(a,4(f12.6,1x))
1883 format(a,4(a11,a1))

    write(errstr,*)'ZZZ post1st amoeba'
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st low       = ',low
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st high      = ',high
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st ndim      = ',ndim
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st parms     = ',parms
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st dparms    = ',dparms
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st mask      = ',mask
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st reslo     = ',reslo
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st reshi     = ',reshi
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st it        = ',it
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st chilo     = ',chilo
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st chihi     = ',chihi
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st gaussind  = ',gauss_ind
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st gaussdep  = ',gauss_dep
    call message(AppMsg,Noisy)
    write(errstr,*)'post1st delta     = ',delta
    call message(AppMsg,Noisy)
    write(errstr,*)''

    call message(AppMsg,Always)
    write(errstr,'(a,3(f9.3),a,3(f9.3))')'1st rl = ',reslo,' / rh = ',reshi
    call message(AppMsg,Always)
    write(errstr,'(2(a,e12.3))')'1st cl = ',chilo,' / ch = ',chihi
    call message(AppMsg,Always)
    allocate(v1(nverls),temp(nverls))
    maxp = 0
    v1 = 0.d0     ! zero out from above
    do i=1,nverls ! replaces IDL where function
       v1(i) = abs(hist_rate(i)-reslo(2))
    end do
    temp  = minval(v1)
    maxi  = findloc(v1,temp(1))
    maxp  = maxi(1)
    imaxp = int(maxp)
    rlow  = max(1.,maxp-delt+1)
    rhigh = min(maxp+delt+1,real(size(hist_rate)))
    low   = int(rlow)
    high  = int(rhigh)
    delta = high - low + 1
    j = 1
    do i=low,high
       gauss_ind(j) = real(hist_rate(int(i)))  ! nint???
       gauss_dep(j) = real(ihist_cnts(int(i))) ! nint???
       kip_x(j)=real(hist_rate(i),8)
       kip_y(j)=real(ihist_cnts(i),8)
       write(errstr,*)"2nd gauss_ind(",j,") = ",gauss_ind(j),&
            "gauss_dep(",j,") = ",gauss_dep(j)
       call message(AppMsg,Verbose)
       j = j + 1
    end do
    parms = (/ real(ihist_cnts(imaxp),8), real(hist_rate(imaxp),8), 0.1D0 /)
    dparms = parms / 10.d0
    write(errstr,1916) &
         '***** 2nd amoeba: low =',low,' peak = ',imaxp,' high = ',high
    call message(AppMsg,Always)
    write(errstr,1916)'***** 2nd amoeba: cnts(low) =',ihist_cnts(low),&
         ' cnts(peak)= ',ihist_cnts(imaxp),' cnts(high) = ',ihist_cnts(high)
    call message(AppMsg,Always)
    write(errstr,'(a,3f9.3)')" 2nd parms = ",parms
    call message(AppMsg,Always)
    !
    ! second iteration of kpamoeba 22222222222
    !
    write(errstr,*)'ZZZ pre2nd amoeba'
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd low       = ',low
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd high      = ',high
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd ndim      = ',ndim
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd parms     = ',parms
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd dparms    = ',dparms
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd mask      = ',mask
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd reslo     = ',reslo
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd reshi     = ',reshi
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd it        = ',it
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd chilo     = ',chilo
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd chihi     = ',chihi
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd gaussind  = ',gauss_ind
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd gaussdep  = ',gauss_dep
    call message(AppMsg,Noisy)
    write(errstr,*)'pre2nd delta     = ',delta
    call message(AppMsg,Noisy)
    write(errstr,*)''
    call message(AppMsg,Noisy)

    kip_parms=parms
    !call kpamoeba(ndim,parms,dparms,mask,0.01d0,100,reslo,reshi,it,chilo,&
    !     chihi,gauss_ind,gauss_dep,delta)

    !print *,"kip old2: ",reslo,chilo
    !treslo = reslo
    !tchilo = chilo
    call nelmin(kip_gauss,ndim,kip_parms,reslo,chilo,0.010d0,dparms,10,100,it,kip_rest,kip_ierr)
    !print *,"kip new2: ",reslo,chilo

    !call diffit(treslo(1),reslo(1),cdif1)
    !call diffit(treslo(2),reslo(2),cdif2)
    !call diffit(treslo(3),reslo(3),cdif3)
    !call diffit(chilo,tchilo,cdif4)

    write(errstr,*)"2nd pass:"
    call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)
    !write(errstr,1882)"old         : ",treslo(1),treslo(2),treslo(3),tchilo
    !call message(AppMsg,Always)
    write(errstr,1882)"new         : ",reslo(1),reslo(2),reslo(3),chilo
    call message(AppMsg,Always)
    !write(errstr,1883)"percent diff: ",cdif1,"%",cdif2,"%",cdif3,"%",cdif4,"%"
    !call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)
    
    write(errstr,*)'ZZZ post2nd amoeba'
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd low       = ',low
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd high      = ',high
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd ndim      = ',ndim
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd parms     = ',parms
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd dparms    = ',dparms
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd mask      = ',mask
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd reslo     = ',reslo
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd reshi     = ',reshi
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd it        = ',it
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd chilo     = ',chilo
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd chihi     = ',chihi
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd gaussind  = ',gauss_ind
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd gaussdep  = ',gauss_dep
    call message(AppMsg,Noisy)
    write(errstr,*)'post2nd delta     = ',delta
    call message(AppMsg,Noisy)
    write(errstr,*)''
    call message(AppMsg,Noisy)
    write(errstr,'(a,3(f9.3),a,3(f9.3))')'2nd rl = ',reslo,' / rh = ',reshi
    call message(AppMsg,Always)
    write(errstr,'(2(a,e12.3))')'2nd cl = ',chilo,' / ch = ',chihi
    call message(AppMsg,Always)

    v1 = 0.d0     ! zero out from above
    temp = 0.d0
    do i=1,nverls ! replaces IDL where function
       v1(i) = abs(hist_rate(i)-reslo(2))
    end do
    temp  = minval(v1)
    maxi  = findloc(v1,temp(1))
    maxp  = maxi(1)
    imaxp = int(maxi(1))
    rlow  = max(1.,maxp-delt+1)
    rhigh = min(maxp+delt+1,real(size(hist_rate)))
    low   = int(rlow)
    high  = int(rhigh)
    delta = high - low + 1
    j = 1
    do i=low,high
       gauss_ind(j) = real(hist_rate(int(i)))  ! nint???
       gauss_dep(j) = real(ihist_cnts(int(i))) ! nint???
       kip_x(j)=real(hist_rate(i),8)
       kip_y(j)=real(ihist_cnts(i),8)
       write(errstr,*)"2nd gauss_ind(",j,") = ",gauss_ind(j),&
            "gauss_dep(",j,") = ",gauss_dep(j)
       call message(AppMsg,Verbose)
       j = j + 1
    end do
    parms = (/ real(ihist_cnts(imaxp),8), real(hist_rate(imaxp),8), 0.1D0 /)
    dparms = parms / 10.d0
    write(errstr,1916) &
         '***** 3rd amoeba: low =',low,' peak = ',imaxp,' high = ',high
    call message(AppMsg,Always)
    write(errstr,1916)'***** 3rd amoeba: cnts(low) =',ihist_cnts(low),&
         ' cnts(peak)= ',ihist_cnts(imaxp),' cnts(high) = ',ihist_cnts(high)
    call message(AppMsg,Always)
    write(errstr,'(a,3f9.3)')" 3rd parms = ",parms
    call message(AppMsg,Always)    
    !
    ! third iteration of kpamoeba 33333333333333
    !
    write(errstr,*)'ZZZ pre3rd amoeba'
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd low       = ',low
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd high      = ',high
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd ndim      = ',ndim
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd parms     = ',parms
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd dparms    = ',dparms
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd mask      = ',mask
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd reslo     = ',reslo
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd reshi     = ',reshi
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd it        = ',it
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd chilo     = ',chilo
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd chihi     = ',chihi
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd gaussind  = ',gauss_ind
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd gaussdep  = ',gauss_dep
    call message(AppMsg,Noisy)
    write(errstr,*)'pre3rd delta     = ',delta
    call message(AppMsg,Noisy)
    write(errstr,*)''
    call message(AppMsg,Noisy)

    kip_parms=parms
    !call kpamoeba(ndim,parms,dparms,mask,0.01d0,100,reslo,reshi,it,chilo,&
    !     chihi,gauss_ind,gauss_dep,delta)

    !print *,"kip old3: ",reslo,chilo
    !treslo = reslo
    !tchilo = chilo
    call nelmin(kip_gauss,ndim,kip_parms,reslo,chilo,0.010d0,dparms,10,100,it,kip_rest,kip_ierr)
    !print *,"kip new3: ",reslo,chilo

    !call diffit(treslo(1),reslo(1),cdif1)
    !call diffit(treslo(2),reslo(2),cdif2)
    !call diffit(treslo(3),reslo(3),cdif3)
    !call diffit(tchilo,chilo,cdif4)


    write(errstr,*)"3rd pass:"
    call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)
    !write(errstr,1882)"old         : ",treslo(1),treslo(2),treslo(3),tchilo
    !call message(AppMsg,Always)
    write(errstr,1882)"new         : ",reslo(1),reslo(2),reslo(3),chilo
    call message(AppMsg,Always)
    !write(errstr,1883)"percent diff: ",cdif1,"%",cdif2,"%",cdif3,"%",cdif4,"%"
    !call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)

    write(errstr,*)'ZZZ post3rd amoeba'
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd low       = ',low
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd high      = ',high
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd ndim      = ',ndim
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd parms     = ',parms
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd dparms    = ',dparms
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd mask      = ',mask
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd reslo     = ',reslo
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd reshi     = ',reshi
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd it        = ',it
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd chilo     = ',chilo
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd chihi     = ',chihi
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd gaussind  = ',gauss_ind
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd gaussdep  = ',gauss_dep
    call message(AppMsg,Noisy)
    write(errstr,*)'post3rd delta     = ',delta
    call message(AppMsg,Noisy)
    write(errstr,*)''
    call message(AppMsg,Noisy)

    write(errstr,'(a,3(f9.3),a,3(f9.3))')'3rd rl = ',reslo,' / rh = ',reshi
    call message(AppMsg,Always)
    write(errstr,'(2(a,e12.3))')'3rd cl = ',chilo,' / ch = ',chihi
    call message(AppMsg,Always)
    !
    allocate(hist_modl(nverls),hist_resi(nverls))

    hist_modl = 0.d0
    hist_resi = 0.d0
    do i=1,nverls
       !this is what was fit
       hist_modl(i) = reslo(1) * exp(-(hist_rate(i)-reslo(2))**2/2./reslo(3)/reslo(3))
       hist_resi(i) = real(ihist_cnts(i),8) - hist_modl(i) !this is the histogram-the fit
       write(errstr,*)' hist_modl(',i,') = ',hist_modl(i),' hist_resi(',i,') = ',hist_resi(i)
       call message(AppMsg,Noisy)
    end do
    !
    ! evaluate the goodness of fit
    !
    ! IDL: plc=where(hist_modl GE 0.00001*max(hist_modl))
    !
    j = 1
    v2 = 0.00001d0 * maxval(hist_modl)
    allocate (plc(nverls))
    plc = 0
    do i=1,nverls
       if (hist_modl(i) .ge. v2) then
          plc(j) = i
          j = j + 1
       end if
    end do
    j = j - 1
    ! bhp: shrink plc
    allocate(plc2(j))
    do i = 1, j
       plc2(i) = plc(i)
    end do
    h_imme = 0.d0
    h_totl = 0.d0
    l_imme = 0.d0
    l_totl = 0.d0
    badh = .FALSE.
    badl = .FALSE.
    do i=1,size(hist_modl)
       write(errstr,*)"hist_modl(",i,") = ",hist_modl(i)
       call message(AppMsg,Noisy)
    enddo
    if(maxval(plc2) .gt. high) then 
       if(high .lt. nverls-1) then
          h_imme = sum(real(ihist_cnts(high+1:min(high+3,nverls-1)),8))/maxval(hist_modl)/3.
          h_totl = sum(hist_resi(high+1:maxval(plc2)))/sum(hist_modl(low:high))
          do i=high-1,nverls
             write(errstr,*)'ihist_cnts(',i,') = ',ihist_cnts(i)
             call message(AppMsg,Noisy)
          end do
       else
          h_imme=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
          h_totl=-9.99d+99 ! be used in this context. Was -1.0
          badh = .TRUE.
       endif
       l_imme=sum(real(ihist_cnts(max(low-3,1):low-1),8))/maxval(hist_modl)/3.
       if(minval(plc2) .lt. low)then 
          l_totl=sum(hist_resi(minval(plc):low-1))/sum(hist_modl(low:high))
       else
          call warning("lowOutOfRange","low outside acceptable range")
          write(errstr,*)"low = ",low," max(plc) = ", &
               maxval(plc2),' L_IMME/L_TOTL set to -9.99E+99'
          call message(AppMsg,Always)
          l_imme=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
          l_totl=-9.99d+99 ! be used in this context. Was -1.0 until 2021 SAS
          badl = .TRUE.
       end if
    else
       call warning("MaxPLCltHigh","Max plc value outside acceptable range")
          write(errstr,*)" max(plc) = ",maxval(plc2)," high = ",high,&
               ' H_IMME/H_TOTL/L_IMME/L_TOTL all set to -9.99E+99'
          call message(AppMsg,Always)
          h_imme=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
          h_totl=-9.99d+99 ! be used in this context. Was -1.0 until 2021 SAS
          l_imme=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
          l_totl=-9.99d+99 ! be used in this context. Was -1.0 until 2021 SAS
          badh = .TRUE.
          badl = .TRUE.
    end if
    write(errstr,2525)'H_IMME = ',h_imme,' H_TOTL = ',h_totl,&
         ' L_IMME = ',l_imme,' L_TOTL = ',l_totl
2525 format(a,g10.3,a,g10.4,a,g10.4,a,g10.4)
    call message(AppMsg,Always)
    ! bhp: diagnosis numbers for when fit eval blows up. 
    if (badh) then 
       write(errstr,*)'Components for calculating H_IMME/H_TOTL may have blown up.'
       call warning("HfitEvalBlownUp",trim(adjustl(errstr)))
       write(errstr,*)'Suggest altering input parameters'
       write(errstr,*)'low, high, maxval(plc2) = ',low,high,maxval(plc2)
       call message(AppMsg,Always)
       write(errstr,*)'maxval(hist_modl) = ',maxval(hist_modl)
       call message(AppMsg,Always)
       write(errstr,*)'sum cnts = ',sum(real(ihist_cnts(high+1:min(high+3,nverls-1)),8))
       call message(AppMsg,Always)
       write(errstr,*)'sum resi = ',sum(hist_resi(high+1:maxval(plc2)))
       call message(AppMsg,Always)
       write(errstr,*)'sum modl = ',sum(hist_modl(low:high))
       call message(AppMsg,Always)
       write(errstr,*)' H_IMME/H_TOTL defaulting to -9.99E+99.'
       call message(AppMsg,Always)
    endif
    if (badl) then 
       write(errstr,*)'Components for calculating L_IMME/L_TOTL may have blown up.'
       call warning("LfitEvalBlownUp",trim(adjustl(errstr)))
       write(errstr,*)'low, high, maxval(plc2) = ',low,high,maxval(plc2)
       call message(AppMsg,Always)
       write(errstr,*)'maxval(hist_modl) = ',maxval(hist_modl)
       call message(AppMsg,Always)
       write(errstr,*)'sum cnts = ',sum(real(ihist_cnts(max(low-3,1):low-1),8))
       call message(AppMsg,Always)
       write(errstr,*)'sum resi = ',sum(hist_resi(minval(plc):low-1))
       call message(AppMsg,Always)
       write(errstr,*)'sum modl = ',sum(hist_modl(low:high))
       call message(AppMsg,Always)
       write(errstr,*)' L_IMME/L_TOTL defaulting to -9.99E+99.'
       call message(AppMsg,Always)
    endif
    !
    !  write out a QDP plot file and fill GTI for filtering
    !  descoped for ratio method in SAS-20.1
    if (method .eq. 1) then
       write(errstr,*) "CLEAN_LC: Writing out plots to QDP file: ",&
            trim(adjustl(histqdpFile))
       call message(AppMsg,Verbose)
       ! bhp: is the below still needed ???!!!
       if((rangescale .gt. 1.99) .and. (rangescale .lt. 2.01)) rangescale = 1.99
       delt = rangescale*sqrt(maxp)*delt/5.0
       hist_cnts = real(ihist_cnts)
       ! KDK additions, 10/2020
       apar(1) = parms(1)
       apar(2) = reslo(2)
       apar(3) = reslo(3)
       ! bhp: what if low .le. 1???!!!
       rlow = (low-1)*binw+bin0
       rhigh = (high-1)*binw+bin0    
       !
       ! pair down rates arrays for plotting
       !
       allocate(rate_lc(binned_lc))
       rate_lc(1:binned_lc) = modrate_lc(1:binned_lc)

       if (specialSmode) then
          allocate(rate_clc(binned_clc))
          rate_clc(1:binned_clc) = modrate_clc(1:binned_clc)
       else
          allocate(rate_clc(1))
       end if
       !
       ! write the data to a QDP file (change to python???!!!)
       !
       call write_qdp_gti(obsid,histqdpFile,gtitxtFile,apar,hist_cnts,&
            tstart_lc,tend_lc,tstart_clc,tend_clc,tstart_flc,tend_flc,&
            binned_lc,binned_clc,binned_flc,rate_lc,rate_clc,&
            bin0,binw,rlow,rhigh,method,ratio,specialSmode,rangescale,&
            norm,width,center,h_imme,h_totl,l_imme,l_totl)
       deallocate (rate_lc)
    end if
    deallocate(new_modrate_lc)
    deallocate(new_modrate_clc)
    deallocate(new_modrate_flc)
    deallocate(modrate_lc)
    deallocate(modrate_clc)
    deallocate(modrate_flc)
    deallocate (tstart_lc,tend_lc)
    deallocate (tstart_clc,tend_clc)
    deallocate (tstart_flc,tend_flc)

       
    return

  end subroutine clean_lc
  !
  subroutine write_qdp_gti(obsid,histqdpF,gtitxtF,apar,hist_cnts,&
       tstart_lc,tend_lc,tstart_clc,tend_clc,tstart_flc,tend_flc,&
       nbins_lc,nbins_clc,nbins_flc,modrate_lc,modrate_clc,&
       bin0,binw,rlow,rhigh,method,ratio,specialSmode,rangescale,&
       norm,width,center,h_imme,h_totl,l_imme,l_totl)
    !
    ! write histo, rate, information to QDP file for later plotting
    ! should be changed to plot directly to PS! 
    !
    character(len=10), intent(in) :: obsid
    character(len=256), intent(in) :: histqdpF, gtitxtF
    integer, intent(in) :: nbins_lc, nbins_clc, nbins_flc, method
    integer :: i, ii, n, nn, lun, pflg = 1, ierr, status
    logical :: specialSmode
    real(kind=double), intent(in) :: apar(3)
    real(kind=double), intent(in), dimension(:) :: hist_cnts
    real(kind=double), intent(in), dimension(:) :: tstart_lc, tend_lc, &
         tstart_clc, tend_clc, tstart_flc, tend_flc, &
         modrate_lc, modrate_clc
    real(kind=double) :: ratelimm, ratelimp, x, dt, y, ys
    real(kind=double), intent(inout) :: norm, width, center
    real(kind=double) :: d1 = -1.0, d2 = 0.0, sca = 1.5, &
         xmax, ymax, ymaxf, yymax
    real(kind=double), allocatable, dimension(:) :: tstart, tend
    real(kind=double) :: rlow, rhigh, rat
    real(kind=double), intent(in) :: h_imme, h_totl, l_imme, l_totl
    real(kind=double), intent(in) :: ratio, rangescale, bin0, binw
    ierr = 0
    status = 0
    lun = GetLun()
    if(status .ne. 0) then ! does this do anything, should be error???
       call warning("NoOpenQDPfile",&
            "WRITE_QDP_GTI: Unable to open QDP file for output.")
    endif

    if( clobber () ) then
       open(unit=lun,status='unknown',form='formatted',&
            file=trim(adjustl(histqdpF)))
    else
       open(unit=lun,status='new',form='formatted',&
            file=trim(adjustl(histqdpF)))
    endif
    write(lun,*) 'viewport 0.05 0.05 0.98 0.89'
    write(lun,*) 'font roman'
    write(lun,*) 'cs 1.0'
    write(lun,*) 'skip sing'
    write(lun,*) 'gap errors'
    Write(lun,*) 'col 1 on 1'
    write(lun,*) 'col 3 on 2'
    write(lun,*) 'col 1 on 3'
    write(lun,*) 'col 3 on 4'
    write(lun,*) 'col 1 on 5'
    write(lun,*) 'col 3 on 6'
    write(lun,*) 'col 2 on 7'
    write(lun,*) 'col 4 on 8'
    write(lun,*) 'line step 2'
    write(lun,*) 'read serr 2'
    write(lun,*) 'read serr 4'
    write(lun,*) 'read serr 5'
    write(lun,*) 'read serr 6'
    write(lun,*) 'read serr 7'
    write(lun,*) 'win 1'
    write(lun,*) 'yplot 1 2'
    write(lun,*) 'loc 0.03 0.72 0.97 1.0'
    write(lun,*) 'lab x Count Rate (counts/s)'
    write(lun,*) 'la y N'
    xmax = 2.0*apar(2)
    if(xmax .lt. 5.0) xmax = 5.0
    write(lun,*) 'r x1 0 ', xmax
    ymax = 0.0
    yymax = 0.0
    do i=1,size(hist_cnts) !!! fix this !!!
       !TBD(KDK)
       ! y = ap(1)*exp(-ap(2)*(i - ap(3))*(i - ap(3)))
       y = apar(1)*exp(-((i-1)*binw+bin0-apar(2))**2/2./apar(3)/apar(3))
       if(y + 1.1*sqrt(y) .gt. yymax) &
            yymax = y + 1.1*sqrt(y)
       if(hist_cnts(i) + 1.1*sqrt(hist_cnts(i)) .gt. ymax) &
            ymax = hist_cnts(i) + 1.1*sqrt(hist_cnts(i))
    enddo
    if((yymax .gt. 0.0) .and. (ymax/yymax .gt. 5)) ymax = yymax
    ymax = 1.1*ymax
    ymaxf = ymax
    write(errstr,*) &
         "WRITE_QDP_GTI: Ymax = ",ymax
    call message(AppMsg,Verbose)
    write(lun,*) 'r y1 0 ', ymax
    write(lun,*) 'la 1 p ', 0.5*xmax, 0.9*ymax,&
         ' "Count Rate Histogram"'
    write(lun,*) 'la 11 p ', 0.02*xmax, 0.9*ymax,&
         ' cs 0.6 just left "Fit Lims: Blue"'
    write(lun,*) 'la 12 p ', 0.02*xmax, 0.8*ymax,&
         ' cs 0.6 just left "Select Lims: Red"'
    write(lun,*) 'la 13 p ', 0.8*xmax, 1.05*ymax,&
         ' cs 0.8 just left "ObsID: ',obsid, '"'
    norm = apar(1)
    width = apar(3)
    center = apar(2)
    write(errstr,'(a,f9.3,a,f9.3,a,f9.3)') &
         "WRITE_QDP_GTI: Norm = ",norm," Width = ",width," Center = ",center
    call message(AppMsg,Always)
    write(lun,3008) 0.78*xmax, 0.9*ymax, norm
3008 format('la 14 p ',f10.4,f10.2,' cs 0.8 just ',&
         'left "Fit Norm: ',f7.1,'"')
    write(lun,3009) 0.78*xmax, 0.8*ymax, width
3009 format('la 15 p ',f10.4,f10.2,' cs 0.8 just ',&
         'left "Fit Width: ',f7.3,' c/s"')
    write(lun,3010) 0.78*xmax, 0.7*ymax, center
3010 format('la 16 p ',f10.4,f10.2,' cs 0.8 just ',&
          'left "Fit Center: ',f7.3,' c/s"')
    
    write(lun,3011) 0.78*xmax, 0.6*ymax, h_imme
3011 format('la 17 p ',f10.4,f10.2,' cs 0.8 just ',&
          'left "h_imme: ',f7.3)
    write(lun,3012) 0.78*xmax, 0.5*ymax, h_totl
3012 format('la 18 p ',f10.4,f10.2,' cs 0.8 just ',&
          'left "h_totl: ',f7.3)
    write(lun,3013) 0.78*xmax, 0.4*ymax, l_imme
3013 format('la 19 p ',f10.4,f10.2,' cs 0.8 just ',&
          'left "l_imme: ',f7.3)
    write(lun,3014) 0.78*xmax, 0.3*ymax, l_totl
3014 format('la 20 p ',f10.4,f10.2,' cs 0.8 just ',&
          'left "l_totl: ',f7.3)
    !
    !  Write out the control lines for the plot of the FOV light curve
    !
    write(errstr,*) &
         "WRITE_QDP_GTI: Write out the control lines for the plot of the FOV light curve"
    call message(AppMsg,Noisy)
    write(lun,*) 'win 2'
    write(lun,*) 'yplot 3 4'
    write(lun,*) 'loc 0.03 0.39 0.97 0.67'
    write(lun,*) 'la x Time (s)'
    write(lun,*) 'la y Count Rate (counts/s)'
    write(lun,*) 'r x2 0 ', tend_lc(nbins_lc) - tstart_lc(1)

    ymax = 0.0
    do i=1,nbins_lc
       dt = tend_lc(i) - tstart_lc(i)
       if(modrate_lc(i) + sqrt(modrate_lc(i)*dt)/dt .gt. ymax) then
          ymax = modrate_lc(i) + sqrt(modrate_lc(i)*dt)/dt
       endif
    enddo
    write(lun,*) 'r y2 0.1 ', ymax
    write(lun,*) 'la 2 p ', 0.5*(tend_lc(nbins_lc)-tstart_lc(1)),& 
         0.5*ymax,' "FOV Light Curve"'
    write(lun,*) 'log y'
    !
    !  Write out the control lines for the plot of the corner light curve
    !
    if (specialSmode) then
       write(errstr,*) &
            "WRITE_QDP_GTI: Write out the control lines for the plot of the Corner light curve"
       call message(AppMsg,Noisy)
       write(lun,*) 'win 3'
       write(lun,*) 'yplot 5 6'
       write(lun,*) 'loc 0.03 0.06 0.97 0.34'
       write(lun,*) 'la x Time (s)'
       write(lun,*) 'la y Count Rate (counts/s)'
       write(lun,*) 'r x3 0 ', tend_clc(nbins_clc)-tstart_clc(1)
       ymax = 0.0
       do i=1,nbins_clc
          dt = tend_clc(i) - tstart_clc(i)
          if(modrate_clc(i) + sqrt(modrate_clc(i)*dt)/dt .gt. ymax) then
             ymax = modrate_clc(i) + sqrt(modrate_clc(i)*dt)/dt
          endif
       enddo
       write(lun,*) 'r y3 0.0 ', ymax
       write(lun,*) 'la 3 p ', 0.5*(tend_clc(nbins_clc)-tstart_clc(1)),&
            0.9*ymax,' "Corner Light Curve"'
    endif
    !
    ! Write out the data for the count rate histogram plot
    !
    do i=1,size(hist_cnts) ! should this be changed??
       !x = 0.02*i - 0.01 incorrect line (KDK)
       x = (i-1)*binw+bin0
       y = apar(1)*exp(-((i-1)*binw+bin0-apar(2))**2/2./apar(3)/apar(3))
       !y = ap(1)*exp(-ap(2)*(i - ap(3))*(i - ap(3)))
       !ap(1) * exp( -ap(2) * (i - ap(3)) * (i - ap(3)) )
       write(lun,2000) x, hist_cnts(i), sqrt(hist_cnts(i)),&
            y, d1, d2, d1, d2, d1, d2, d1, d2, d1, d1
!       write(50,*)hist_cnts(i),' ',x,' ',y
    enddo
2000 format(1x,f14.3,3f10.3,10f5.1)

    !
    !  Set up the good time intervals
    !
    write(errstr,*) "WRITE_QDP_GTI: Set up GTIs"
    call message(AppMsg,Noisy)
    !TBD(KDK)
    !ratelimm = 0.02*(ap(3) - sca/sqrt(ap(2)))
    !ratelimp = 0.02*(ap(3) + sca/sqrt(ap(2)))
    ratelimm = apar(2) - rangescale*apar(3)
    ratelimp = apar(2) + rangescale*apar(3)
    write(errstr,*) "WRITE_QDP_GTI: LimL = ",ratelimm," LimH = ",ratelimp
    call message(AppMsg,Noisy)
    write(errstr,*) "WRITE_QDP_GTI: Center = ",apar(2),&
         " Width = ",apar(3)," Scale = ",rangescale
    call message(AppMsg,Noisy)
    write(errstr,*) "WRITE_QDP_GTI: RLow = ",rlow," RHigh = ",rhigh
    call message(AppMsg,Noisy) 
    nn = 0
    allocate(tstart(nbins_lc))
    allocate(tend(nbins_lc))
    do i=1,nbins_lc
       if((modrate_lc(i) .gt. 0.0) .and. (modrate_lc(i) .le. ratelimp) &
            .and. (modrate_lc(i) .ge. ratelimm)) then
          nn = nn + 1
          tstart(nn) = tstart_lc(i)
          tend(nn) = tend_lc(i)
       endif
    enddo
    !
    !  If the intervals are close then run them together
    !
    n = 1
    do i=2,nn
       if(abs(tstart(i) - tend(i-1)) .gt. 5.0) then
          tend(n) = tend(i-1)
          n = n + 1
          tstart(n) = tstart(i)
       endif
    enddo
    if (nn.eq.0) then
       nn = 1   ! singular matrix solution
       call warning("SingularMatrix","No Count Rate Histogram Produced")
    end if

    if (nn.lt.0) call warning("nnLTzero","Close interval running problem. NN is LT 0???")
    tend(n) = tend(nn)
    !
    !  Eliminate the short time intervals
    !
    nn = 0
    do i=1,n
       if(tend(i)-tstart(i) .gt. 100.0) then
          nn = nn + 1
          tstart(nn) = tstart(i)
          tend(nn) = tend(i)
       endif
    enddo
    !
    !  Write out a count rate plot file
    !
    write(errstr,*) "WRITE_QDP_GTI: Writing out a FOV count rate plot file."
    call message(AppMsg,Noisy)

    if(pflg .eq. 1) then
       do i=1,nbins_lc
          x = (tstart_lc(i) + tend_lc(i))/2.0
          dt = tend_lc(i) - tstart_lc(i)
          y = modrate_lc(i)
          ys = sqrt(modrate_lc(i)*dt)/dt
          write(lun,2010) x-tstart_lc(1), d1, d2, d1,& 
               y, ys, d1, d2, d1, d2, d1, d2, d1, d1
2010      format(1x,f14.3,3f5.1,2f10.3,8f5.1)
       enddo
       do i=1,nbins_lc
          x = (tstart_lc(i) + tend_lc(i))/2.0
          do ii=1,nn
             if((x .ge. tstart(ii)) .and. (x .le. tend(ii))) then
                dt = tend_lc(i) - tstart_lc(i)
                y = modrate_lc(i)
                ys = sqrt(modrate_lc(i)*dt)/dt
                write(lun,2020) x-tstart_lc(1), d1, d2, d1,&
                     d1, d2, y, ys, d1, d2, d1, d2, d1, d1
2020            format(1x,f14.3,5f5.1,2f10.3,6f5.1)
             endif
          enddo
       enddo
       !
       !  Plot the corner data
       !
       if (specialSmode) then
          write(errstr,*) &
               "WRITE_QDP_GTI: Writing out a CORNER count rate plot file."
          call message(AppMsg,Noisy)
          do i=1,nbins_clc
             x = (tstart_clc(i) + tend_clc(i))/2.0
             dt = tend_clc(i) - tstart_clc(i)
             y = modrate_clc(i)
             ys = sqrt(modrate_clc(i)*dt)/dt
             write(lun,2030) x-tstart_clc(1), d1, d2, d1,&
                  d1, d2, d1, d2, y, ys, d1, d2, d1, d1 
2030         format(1x,f14.3,7f5.1,2f10.3,4f5.1)
          enddo

          do i=1,nbins_clc
             x = (tstart_clc(i) + tend_clc(i))/2.0
             do ii=1,nn
                if((x .ge. tstart(ii)) .and. (x .le. tend(ii))) then
                   dt = tend_clc(i) - tstart_clc(i)
                   y = modrate_clc(i)
                   ys = sqrt(modrate_clc(i)*dt)/dt
                   write(lun,2040) x-tstart_clc(1), d1, d2, d1,&
                        d1, d2, d1, d2, d1, d2, y, ys, d1, d1
2040               format(1x,f14.3,9f5.1,2f10.3,2f5.1)
                endif
             enddo
          enddo
       endif
    endif
    ! write the selection limits (red lines) 
    write(lun,2050) ratelimm, d1, d2, d1,& 
         d1, d2, d1, d2, d1, d2, d1, d2, d2, d1
    write(lun,2050) ratelimm, d1, d2, d1,& 
         d1, d2, d1, d2, d1, d2, d1, d2, 1.05*ymaxf, d1
    write(lun,2050) ratelimp, d1, d2, d1,&
         d1, d2, d1, d2, d1, d2, d1, d2, 1.05*ymaxf, d1
    write(lun,2050) ratelimp, d1, d2, d1,&
         d1, d2, d1, d2, d1, d2, d1, d2, d2, d1
2050 format(1x,f14.3,11f5.1,f10.3,f5.1)
    ! write the fit limits (blue lines)
    write(lun,2060) rlow, d1, d2, d1,&
         d1, d2, d1, d2, d1, d2, d1, d2, d1, d2
    write(lun,2060) rlow, d1, d2, d1,& 
         d1, d2, d1, d2, d1, d2, d1, d2, d1, 1.05*ymaxf
    write(lun,2060) rhigh, d1, d2, d1,& 
         d1, d2, d1, d2, d1, d2, d1, d2, d1, 1.05*ymaxf
    write(lun,2060) rhigh, d1, d2, d1,&
         d1, d2, d1, d2, d1, d2, d1, d2, d1, d2
2060 format(1x,f14.3,12f5.1,f10.3)
    close(lun)
    !
    !  Write out an ascii file of the gtis
    !
    status = 0
    call ftgiou(lun,status)
    if(status .ne. 0) then
       write(errstr,*) "WRITE_QDP_GTI: Unable to write to GTI ASCII file."
       call message(AppMsg,Verbose)
    endif
    if( clobber() ) then
       open(lun,status='unknown',form='formatted',file=trim(adjustl(gtitxtF)))
    else
       open(lun,status='new',form='formatted',file=trim(adjustl(gtitxtF)))
    endif
    do i=1,nn
       write(lun,*) tstart(i), tend(i), '  +'
    enddo
    close(lun)
    
    return
  end subroutine write_qdp_gti
  !
  ! create common eventlist, image, and lightcurve products
  !
  subroutine makeevimlc (inEVLI_F,method,cornerdef,anndef,inst,&
       expid,smode,revolut,elow,ehigh,patt,fovimg_F,fovlc_F,&
       corevli_F,corimg_F,corlc_F,annEVLI_F,annimg_F,annlc_F,withlong)
    !
    ! evselect runs on event list. should be replaced by selectlib after SAS19!!!
    ! also, add elow and ehigh atributes to header if present
    !
    character(len=2) :: patt, inst
    character(len=5) :: cminchan, cmaxchan
    character(len=10) :: expid, type, suffix, method
    character(len=32) :: flagg, flagf
    character(len=256) :: inEVLI_F, fovimg_F, fovlc_F, corevli_F, &
         corimg_F, corlc_F, annEVLI_F, annimg_F, annlc_F
    character(len=1024) :: command, cornerdef, anndef
    integer :: revolut, elow, ehigh, status
    logical :: smode, warndie, withlong
    write(cminchan,'(I5)') elow
    write(cmaxchan,'(I5)') ehigh
    cminchan = adjustl(cminchan)
    cmaxchan = adjustl(cmaxchan)
    !
    if (inst .eq. 'PN') then ! new for SAS-19
       flagg =  '(#XMMEA_EP)'
       flagf =  '((FLAG & 0xfb0000)==0)'
    else
       flagg = "((FLAG & 0x766aa000)==0)"
       flagf = "(#XMMEA_EM)"
    endif

    write(errstr,*) "Running evselect to create *ALL IMAGE*"
    call message(AppMsg,Verbose)
    ! this isn't actually used for anything. Just a quicklook product ??? 
    ! OLDNEW: fovimg_F = mos1S001-fovimg.FIT/mos1S001-fovim.fits
    type = "allim" ! fov + image (full file)
    suffix = "fits"
    call make_filename(inst,expid,type,elow,ehigh,suffix,fovimg_F,withlong)
    write(errstr,*) "*ALL IMAGE* = "//&
         trim(adjustl(fovimg_F))
    call message(AppMsg,Always)
    command = "evselect table="// trim(adjustl(inEVLI_F)) //" withimageset=yes "//&
         " imageset="// trim(adjustl(fovimg_F)) // " xcolumn=DETX ycolumn=DETY "//&
         "imagebinning=imageSize ximagesize=600 yimagesize=600"
    write(errstr,*) "*ALL IMAGE evselect command = "
    call message(AppMsg,Noisy)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Noisy)
    call system(trim(adjustl(command)))
    call addlohi(fovimg_F,elow,ehigh)
    ! 
    write(errstr,*) "Running evselect to create *FOV LIGHTCURVE*"
    call message(AppMsg,Always)
    ! OLDNEW: fovlc_F = mos1S001-fovlc.FIT/mos1S001-fovlc.fits
    type = "fovlc" ! fov + lightcurve 
    suffix = "fits"
    call make_filename(inst,expid,type,elow,ehigh,suffix,fovlc_F,withlong)
    write(errstr,*) "*FOV LIGHTCURVE * = "//&
         trim(adjustl(fovlc_F))
    call message(AppMsg,Always)
    command = "evselect table="// trim(adjustl(inEVLI_F)) //":EVENTS "//&
         "withfilteredset=yes expression='(PATTERN<="// &
         trim(adjustl(patt)) //")&&(PI in "//&
         "["// trim(adjustl(cminchan)) // ":" // trim(adjustl(cmaxchan)) // "])&&"//&
         trim(adjustl(flagf))//"' filtertype=expression "//&
         "keepfilteroutput=no withrateset=yes rateset="// trim(adjustl(fovlc_F)) //&
         " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
         "timebinsize=1 withtimeranges=no"
    write(errstr,*) "*FOV LIGHTCURVE evselect command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(adjustl(command)))
    call addlohi(fovlc_F,elow,ehigh)
    !
    ! create corner files
    !
    if (smode) then
       ! CORNER EVENTLIST
       write(errstr,*) "Running evselect to create *CORNER EVENTLIST*"
       call message(AppMsg,Always)
       ! OLDNEW corevli_F = mos1S001-corevli.FIT/mos1S001-corev.fits
       type = "corev"
       suffix = "fits"
       call make_filename(inst,expid,type,elow,ehigh,suffix,corevli_F,withlong)
       write(errstr,*) "*CORNER EVENTLIST * = "//&
            trim(adjustl(corevli_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(inEVLI_F)) //":EVENTS " //&
            "withfilteredset=yes expression='(PATTERN<="// trim(adjustl(patt)) //&
            ")&&"//trim(adjustl(flagg))//"&&"//&
            trim(adjustl(cornerdef))//"' filteredset="//&
            trim(adjustl(corevli_F))//&
            " filtertype=expression keepfilteroutput=yes "//&
            "updateexposure=yes filterexposure=yes" 
       write(errstr,*) "*CORNER EVENTLIST evselect command = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(corevli_F,elow,ehigh)
       ! CORNER IMAGE
       write(errstr,*) "Running evselect to create *CORNER IMAGE*"
       call message(AppMsg,Always)
       ! OLDNEW: corimg_F = mos1S001-corimg.FIT/mos1S001-corim.fits
       type = "corim"
       suffix = "fits"
       call make_filename(inst,expid,type,elow,ehigh,suffix,corimg_F,withlong)
       write(errstr,*) "*CORNER IMAGE * = "//&
            trim(adjustl(corimg_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(corevli_F)) //&
            " withimageset=yes imageset="// trim(adjustl(corimg_F)) //&
            " xcolumn=DETX ycolumn=DETY "//&
            "imagebinning=imageSize ximagesize=600 yimagesize=600"
       write(errstr,*) "*CORNER IMAGE evselect command = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(corimg_F,elow,ehigh)
       ! CORNER LIGHTCURVE
       write(errstr,*) "Running evselect to create *CORNER LIGHTCURVE*"
       call message(AppMsg,Always)
       ! OLDNEW: corlc_F = mos1S001-corlc.FIT/mos1S001-corlc.fits
       type = "corlc"
       suffix = "fits"
       call make_filename(inst,expid,type,elow,ehigh,suffix,corlc_F,withlong)
       write(errstr,*) "*CORNER LIGHTCURVE * = "//&
            trim(adjustl(corlc_F))
       call message(AppMsg,Always)
       !
       command = "evselect table="// trim(adjustl(inEVLI_F)) //":EVENTS "//&
            "withfilteredset=yes expression="//&
            "'(PATTERN<="// trim(adjustl(patt)) //")&&(PI in "//&
            "["// trim(adjustl(cminchan)) // ":" // trim(adjustl(cmaxchan)) // "])&&"//&
            trim(adjustl(flagg))//"&&"// trim(adjustl(cornerdef)) //&
            "' filtertype=expression keepfilteroutput=no "//&
            "withrateset=yes rateset="// trim(adjustl(corlc_F)) //&
            " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
            "timebinsize=1 withtimeranges=no"
       write(errstr,*) "*CORNER LIGHTCURVE evselect command = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(corlc_F,elow,ehigh)
    else
       write(errstr,*) "PN SMALL window data, so no corner data *CORNER LIGHTCURVE* created"
       call message(AppMsg,Always)
    endif
    !
    ! create FOV annulus eventlist and lightcurve only for ratio method
    !
    if (method .eq. 'ratio') then
       write(errstr,*) "Running evselect to create *FOV annulus EVENTLIST*"
       call message(AppMsg,Always)
       ! OLDNEW:  fovannevF = mos1S001-annev.FIT" mos1S001-annev.fits
       type = "annev"  
       suffix = "fits" 
       call make_filename(inst,expid,type,elow,ehigh,suffix,annEVLI_F,withlong)
       command = "evselect table="// trim(adjustl(inEVLI_F)) //":EVENTS" //&
            " withfilteredset=yes expression='(PATTERN<="// trim(adjustl(patt))// &
            ")&&"//trim(adjustl(flagf))//"&&"// trim(adjustl(anndef))//"' " //&
            "filteredset="// trim(adjustl(annEVLI_F)) //&
            " filtertype=expression keepfilteroutput=yes"//&
            " updateexposure=yes filterexposure=yes" 
       write(errstr,*) "*FOV annulus eventlist evselect command (ratio method) = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(annEVLI_F,elow,ehigh)
       !
       !
       write(errstr,*) "Running evselect to create *FOV annulus IMAGE*"
       call message(AppMsg,Always)
       ! FOV annulus IMAGE (intermediary/diagnostic) file
       ! OLDNEW:  fovannevF = ???/mos1S001-annim.fits
       type = "annim"  ! or fov-annevc? or whatever
       suffix = "fits" ! 
       call make_filename(inst,expid,type,elow,ehigh,suffix,annimg_F,withlong)
       command = "evselect table="// trim(adjustl(annEVLI_F)) //" withimageset=yes "//&
            " imageset="// trim(adjustl(annimg_F)) // " xcolumn=DETX ycolumn=DETY "//&
            "imagebinning=imageSize ximagesize=600 yimagesize=600"
       write(errstr,*) "*FOV annulus image evselect command (ratio method) = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(annimg_F,elow,ehigh)
       !
       ! make FOV annulus LC in the selected range
       !
       write(errstr,*) "Running evselect to create "//&
            "*FOV annulus LIGHTCURVE (ratio method)*"
       call message(AppMsg,Verbose)
       ! OLDNEW:  fovannlcF = mos1S001-annlc.FIT/mos1S001-annlc.fits
       type = "annlc"
       suffix = "fits" 
       call make_filename(inst,expid,type,elow,ehigh,suffix,annlc_F,withlong)
       write(errstr,*) "*FOV annulus LIGHTCURVE FILE* = "//&
            trim(adjustl(annlc_F))
       call message(AppMsg,Always)
       command = "evselect table="// trim(adjustl(inEVLI_F)) //":EVENTS "//&
            "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(patt)) //&
            ")&&(PI in "//&
            "["// trim(adjustl(cminchan)) // ":" // trim(adjustl(cmaxchan)) // "])&&"//&
            trim(adjustl(flagf))//"&&"// trim(adjustl(anndef)) //"' " //&
            "filtertype=expression keepfilteroutput=no "//&
            "withrateset=yes rateset="// trim(adjustl(annlc_F)) //&
            " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
            "timebinsize=1 withtimeranges=no"
       write(errstr,*) "*FOV annulus LIGHTCURVE evselect command = "
       call message(AppMsg,Noisy)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Noisy)
       call system(trim(adjustl(command)))
       call addlohi(annlc_F,elow,ehigh)       
    end if
    return

  end subroutine makeevimlc
  !
  ! define the selection expressions for evselect
  !
  subroutine defineexpr (method,patt,inst,cornerdef,anndef)

    character(len=2), intent(in) :: inst, patt
    character(len=10), intent(in) :: method
    character(len=1024), intent(inout) :: anndef, cornerdef
    cornerdef = ''
    anndef = ''
    !    
    if (inst .eq. 'M1') then                ! EMOS1
       write (errstr,*) "Defining corner selection expression for EMOS1"
       call message(AppMsg,Noisy)
       cornerdef = "!(CIRCLE(100,-200,17700,DETX,DETY)"//&
            "||CIRCLE(834,135,17100,DETX,DETY)"//&
            "||CIRCLE(770,-803,17100,DETX,DETY)"//&
            "||BOX(-20,-17000,6500,500,0,DETX,DETY)"//&
            "||BOX(5880,-20500,7500,1500,10,DETX,DETY)"//&
            "||BOX(-5920,-20500,7500,1500,350,DETX,DETY)"//&
            "||BOX(-20,-20000,5500,500,0,DETX,DETY)"//&
            "||BOX(-12900,16000,250,4000,0,DETX,DETY)"//&
            "||BOX(80,18600,150,1300,0,DETX,DETY)"//&
            "||BOX(-10,-18800,125,1500,0,DETX,DETY))"
       if (method .eq. "ratio") then 
          write(errstr,*) &
               "Defining FOV annulus selection expression for EMOS1 (ratio method)"
          call message(AppMsg,Noisy)
          anndef = "(CIRCLE(100,-200,16900,DETX,DETY)" //&
               "&&!CIRCLE(100,-200,13900,DETX,DETY))" ! the donut
       end if
    else if (inst .eq. 'M2') then ! EMOS2
       write (errstr,*) "Defining corner selection expression for EMOS2"
       call message(AppMsg,Noisy)
       cornerdef = "!(CIRCLE(435,1006,17100,DETX,DETY)" //&
            "||CIRCLE(-34,68,17700,DETX,DETY)" //&
            "||BOX(-20,-17000,6500,500,0,DETX,DETY)" //&
            "||BOX(5880,-20500,7500,1500,10,DETX,DETY)" //&
            "||BOX(-5920,-20500,7500,1500,350,DETX,DETY)" //&
            "||BOX(-20,-20000,5500,500,0,DETX,DETY))"
       if (method .eq. "ratio") then
          write(errstr,*) &
               "Defining FOV annulus selection expression for EMOS2 (ratio method)"
          call message(AppMsg,Noisy)
          anndef = "(CIRCLE(0,-200,16900,DETX,DETY)" //&
               "&&!CIRCLE(0,-200,13900,DETX,DETY))" ! the donut
       endif
    else
       !
       ! For the PN fin/fout selection
       ! Steve's fout = cornerdef = corners of instrument defined
       !
       write (errstr,*) "Defining corner selection expression for EMPN"
       call message(AppMsg,Noisy)
       cornerdef = "!((DETX,DETY) in circle(-2200,-1100,18080))"
       !
       ! fin = fovanndef == field of view annulus defined
       !
       if (method .eq. "ratio") then
          write(errstr,*) &
               "Defining FOV annulus selection expression for EPN (ratio method)"
          call message(AppMsg,Noisy)
          anndef = "((DETX,DETY) in circle(-2200,-1100,18080))" //&
               "&&!((DETX,DETY) in circle(-2200,-1100,8200))&&"//&
               "(PATTERN <= "// trim(adjustl(patt)) //")&&(((FLAG & 0x762a097c) == 0)"//&
               "||((FLAG & 0x762a017c) == 0))"
          ! this is the ^^^^ 'donut'/annulus
       endif

    end if

    return
    
  end subroutine defineexpr
  
  subroutine addfitpars(in_F,norm,width,center,h_imme,h_totl,l_imme,l_totl)
    !
    ! add fit params to a file's primary header 
    ! 4 = 'open MODIFY'
    ! warndie = fail if can't add attribute
    !
    type(DataSetT) :: in_S
    character (len=256) :: in_F
    integer :: status=0
    real(kind=double) :: norm, width, center, h_imme, h_totl, l_imme, l_totl
    logical :: warndie = .TRUE., badd
    badd = .FALSE.
    write(errstr,*)"Adding fit quality parameters to output file: ",trim(adjustl(in_F))
    call message(AppMsg,Always)
    if (.not.ieee_is_finite(h_imme)) then
       write(errstr,*)"H_IMME has divide by 0, resetting to -9.99E+99 to allow write to FITS attribute."
       call message(AppMsg,Always)
       h_imme=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
       badd = .TRUE.
    endif
    if (.not.ieee_is_finite(h_totl)) then
       write(errstr,*)"H_TOTL has divide by 0, resetting to -9.99E+99 to allow write to FITS attribute."
       call message(AppMsg,Always)
       h_totl=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
       badd = .TRUE.
    endif
    if (.not.ieee_is_finite(l_imme)) then
       write(errstr,*)"L_IMME has divide by 0, resetting to -9.99E+99 to allow write to FITS attribute."
       call message(AppMsg,Always)
       l_imme=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
       badd = .TRUE.
    endif
    if (.not.ieee_is_finite(l_totl)) then
       write(errstr,*)"L_TOTL has divide by 0, resetting to -9.99E+99 to allow write to FITS attribute."
       call message(AppMsg,Always)
       l_totl=-9.99d+99 ! null value since keywords can't be NaN nor can TNULL
       badd = .TRUE.
    endif
    call check_open_set(trim(in_F),in_S,4,status,warndie)
    call setAttribute(in_S,"NORM",norm,"","Fit Normalization")
    call setAttribute(in_S,"WIDTH",width,"c/s","Fit Width")
    call setAttribute(in_S,"CENTER",center,"c/s","Fit Center")
    call setAttribute(in_S,"H_IMME",h_imme,"","meas-fit, high side, close bin")
    call setAttribute(in_S,"H_TOTL",h_totl,"","meas-fit, high side, total bin")
    call setAttribute(in_S,"L_IMME",l_imme,"","meas-fit, low side, close bin")
    call setAttribute(in_S,"L_TOTL",l_totl,"","meas-fit, low side, total bin")
    call release(in_S)
    if (badd) then
       call warning("BadFittingParam",&
            "Your choices for input params are causing gaussian fit to fail. Try tweaking smooth, et al, for better results")
    endif
    return
  end subroutine addfitpars

  real(kind=double) function kip_gauss(parms)

    real(kind=double)::kip_res(kip_n),parms(3)

    parms=abs(parms) ! because the simplex may "find" negative values
    kip_res=parms(1)*exp(-(kip_x-parms(2))**2/2./parms(3)/parms(3))
    kip_gauss=sum((kip_y-kip_res)**2)/size(kip_res)
    !print *,kip_x(1),kip_x(kip_n),kip_y(1),kip_y(kip_n),kip_res(1),kip_res(kip_n)
    !print *,parms(1),parms(2),parms(3),kip_gauss

  end function kip_gauss

  subroutine diffit(val1,val2,cdiff)

    character(len=8) :: cdiff
    real(kind=double) :: val1, val2, diff
    
    diff = abs(val1-val2)/((val1+val2)/2) * 100.
    write(cdiff,"(f8.5)") diff

    return

  end subroutine diffit
    
end module espfilt_mod

